""">>> help(Rect)"""

from buzzard._env import env
import numpy as np
import affine

class Rect(object):
    """Private tool class used to assess the attributes of an imperfect rectangle"""

    def __init__(self, tl, bl, br, tr):
        self.tl = np.float64(tl)
        self.bl = np.float64(bl)
        self.br = np.float64(br)
        self.tr = np.float64(tr)

    @property
    def empty(self):
        return np.all(self.tl == self.tr) or np.all(self.tl == self.bl)

    @property
    def lrvec(self):
        return self.tr - self.tl

    @property
    def horizontal_parallelism_error(self):
        return (self.tr - self.tl) - (self.br - self.bl)

    @property
    def tbvec(self):
        return self.bl - self.tl

    @property
    def vertical_parallelism_error(self):
        return (self.bl - self.tl) - (self.br - self.tr)

    @property
    def coords(self):
        return np.asarray(
            [self.tl, self.bl, self.br, self.tr]
        )

    @property
    def angle(self):
        lrvec = self.lrvec
        return float(np.arctan2(lrvec[1], lrvec[0]) * 180. / np.pi)

    @property
    def size(self):
        lrvec = self.tr - self.tl
        angle = float(np.arctan2(lrvec[1], lrvec[0]) * 180. / np.pi)
        diagvec = self.br - self.tl
        return np.abs(~affine.Affine.rotation(angle) * diagvec, dtype=np.float64)

    @property
    def spatial_precision(self):
        largest_coord = np.abs(self.coords).max().clip(1, np.inf)
        return largest_coord * 10 ** -env.significant

    @property
    def slack_right(self):
        return self.lrvec / self.size[0] * self.spatial_precision

    @property
    def slack_bottom(self):
        return self.tbvec / self.size[1] * self.spatial_precision

    @property
    def tr_slack_angles(self):
        slackr = self.slack_right
        slackb = self.slack_bottom
        return np.asarray((
            _angle_between(self.tl + slackb, self.tr, self.br - slackr),
            _angle_between(self.tl - slackb, self.tr, self.br + slackr),
        ))

    @property
    def tl_slack_angles(self):
        slackr = self.slack_right
        slackb = self.slack_bottom
        return np.asarray((
            _angle_between(self.bl + slackr, self.tl, self.tr + slackb),
            _angle_between(self.bl - slackr, self.tl, self.tr - slackb),
        ))

    @property
    def bl_slack_angles(self):
        slackr = self.slack_right
        slackb = self.slack_bottom
        return np.asarray((
            _angle_between(self.br - slackb, self.bl, self.tl + slackr),
            _angle_between(self.br + slackb, self.bl, self.tl - slackr),
        ))

    @property
    def br_slack_angles(self):
        slackr = self.slack_right
        slackb = self.slack_bottom
        return np.asarray((
            _angle_between(self.tr - slackr, self.br, self.bl - slackb),
            _angle_between(self.tr + slackr, self.br, self.bl + slackb),
        ))

    @property
    def ratio(self):
        return np.linalg.norm(self.lrvec) / np.linalg.norm(self.tbvec)

    @property
    def ratio_slaking(self):
        tbvec0 = (
            self.tbvec / np.linalg.norm(self.tbvec) *
            (np.linalg.norm(self.tbvec) + self.slack_bottom)
        )
        tbvec1 = (
            self.tbvec / np.linalg.norm(self.tbvec) *
            (np.linalg.norm(self.tbvec) - self.slack_bottom)
        )
        lrvec0 = (
            self.lrvec / np.linalg.norm(self.lrvec) *
            (np.linalg.norm(self.lrvec) + self.slack_right)
        )
        lrvec1 = (
            self.lrvec / np.linalg.norm(self.lrvec) *
            (np.linalg.norm(self.lrvec) - self.slack_right)
        )
        ratios = [
            np.linalg.norm(lrvec0) / np.linalg.norm(tbvec0),
            np.linalg.norm(lrvec1) / np.linalg.norm(tbvec0),
            np.linalg.norm(lrvec0) / np.linalg.norm(tbvec1),
            np.linalg.norm(lrvec1) / np.linalg.norm(tbvec1),
        ]
        return np.min(ratios), np.max(ratios)

    def scale(self, rsize):
        aff = ~affine.Affine.rotation(self.angle)
        tl = np.asarray(aff * self.tl)
        br = np.asarray(aff * self.br)
        return np.asarray((br - tl) / rsize, dtype=np.float64)

    def significant_min(self, smallest_reso):
        largest_coord = np.abs(self.coords).max().clip(1, np.inf)
        ratio_pixel_increment = smallest_reso / largest_coord
        return -np.log10(ratio_pixel_increment)

    def abstract_grid_density(self, smallest_reso):
        pixel_precision = self.spatial_precision / smallest_reso
        return np.floor(1 / pixel_precision)

def _angle_between(a, b, c):
    return np.arccos(np.dot(
        (a - b) / np.linalg.norm(a - b),
        (c - b) / np.linalg.norm(c - b),
    )) / np.pi * 180.

def _round_sig(arr, sig_digit):
    assert 0 < sig_digit <= 16
    arr = np.asarray(arr)
    nonzero_mask = arr != 0
    out = np.zeros_like(arr)
    if nonzero_mask.any():
        nz_arr = arr[nonzero_mask]
        sign = np.sign(nz_arr)
        nz_arr *= sign
        round_place = 10 ** (np.ceil(np.log10(nz_arr)) - sig_digit)
        out[nonzero_mask] = round_place * np.around(nz_arr / round_place) * sign
    return out



from buzzard._actors.message import Msg

class ActorComputationAccumulator(object):
    """Actor that takes care of accumulating computed slices needed
    to write 1 cache tile

    TODO Idea:
    If a computation tile A overlap with several cache tiles (a, b, c, d),
    when slicing the numpy array of A into chunks (a, b, c, d), no copy is performed,
    the 4 arrays internally point to A.
    The pointer to A is only released when all 4 cache footprints have been merged.
    In some cases, it might be a good choice to duplicate the slices of A to release
    memory.

    """

    def __init__(self, raster):
        self._raster = raster
        self._alive = True
        self._cache_tiles_accumulations = {}
        self.address = '/Raster{}/ComputationAccumulator'.format(self._raster.uid)

    @property
    def alive(self):
        return self._alive

    # ******************************************************************************************* **
    def receive_combine_this_array(self, compute_fp, array):
        msgs = []

        for cache_fp in self._raster.cache_fps_of_compute_fp[compute_fp]:

            # Fetch and update storage for that cache_fp
            if cache_fp in self._cache_tiles_accumulations:
                store = self._cache_tiles_accumulations[cache_fp]
            else:
                store = {
                    'missing': set(self._raster.compute_fps_of_cache_fp[cache_fp]),
                    'ready': {},
                }
                self._cache_tiles_accumulations[cache_fp] = store
            assert compute_fp in store['missing']
            store['missing'].remove(compute_fp)

            compute_fp_part = compute_fp & cache_fp
            # TODO Idea: Should cache_fp be dilated before the above intersection? This could be a
            #   parameters in facade constructor.
            #   This means also depending on computation_fp that only touch in the border.
            #   On the other hand overlap is a `computation concern`, overlap may not be a `merge concern`
            slices = compute_fp_part.slice_in(compute_fp)
            assert compute_fp_part not in store['ready']
            store['ready'][compute_fp_part] = array[slices]

            # Send news to merger
            if len(store['missing']) == 0:
                msgs += [
                    Msg('Merger', 'merge_those_arrays', cache_fp, store['ready'])
                ]
                del self._cache_tiles_accumulations[cache_fp]
        return msgs

    def receive_die(self):
        """Receive message: The raster was killed"""
        assert self._alive
        self._alive = False
        self._cache_tiles_accumulations.clear()
        self._raster = None
        return []

    # ******************************************************************************************* **



import numpy as np
from osgeo import osr

from buzzard import srs
from buzzard._footprint import Footprint

class BackDatasetConversionsMixin(object):
    """Private mixin for the Dataset class containing the spatial coordinates
    conversion subroutines"""

    def __init__(self, wkt_work, wkt_fallback, wkt_forced, analyse_transformation, **kwargs):

        if wkt_work is not None:
            sr_work = osr.SpatialReference(wkt_work)
        else:
            sr_work = None
        if wkt_fallback is not None:
            sr_fallback = osr.SpatialReference(wkt_fallback)
        else:
            sr_fallback = None
        if wkt_forced is not None:
            sr_forced = osr.SpatialReference(wkt_forced)
        else:
            sr_forced = None

        self.wkt_work = wkt_work
        self.wkt_fallback = wkt_fallback
        self.wkt_forced = wkt_forced
        self.sr_work = sr_work
        self.sr_fallback = sr_fallback
        self.sr_forced = sr_forced
        self.analyse_transformations = analyse_transformation
        super(BackDatasetConversionsMixin, self).__init__(**kwargs)

    def get_transforms(self, sr_virtual, rect, rect_from='virtual'):
        """Retrieve the `to_work` and `to_virtual` conversion functions.

        Parameters
        ----------
        sr_virtual: osr.SpatialReference
        rect: Footprint or extent or None
        rect_from: one of ('virtual', 'work')
        """
        assert rect_from in ['virtual', 'work']

        if self.sr_work is None:
            return None, None

        assert sr_virtual is not None

        to_work = osr.CreateCoordinateTransformation(sr_virtual, self.sr_work).TransformPoints
        to_virtual = osr.CreateCoordinateTransformation(self.sr_work, sr_virtual).TransformPoints

        to_work = self._make_transfo(to_work)
        to_virtual = self._make_transfo(to_virtual)

        if self.analyse_transformations:
            if rect_from == 'virtual':
                an = srs.Analysis(to_work, to_virtual, rect)
            else:
                an = srs.Analysis(to_virtual, to_work, rect)
            if rect is None:
                pass
            elif isinstance(rect, Footprint):
                if not an.ratio_valid:
                    s = ('Error while checking if on-the-fly reprojection could be performed '
                         'between between Dataset\'s `sr_work` and raster\'s `sr_virtual`. '
                         'Cause: {} (The reprojection is too lossy).\n'
                         'Solutions:\n'
                         '- Pass those tests by reducing `buzz.env.significant`.\n'
                         '- First reproject the raster (using `gdalwarp` for example).\n'
                         '- Bypass those tests by passing `analyse_transformation=False` to the '
                         'Dataset\'s constructor.\n'
                         '- Use smaller coordinates.\n'
                         '- Use different spatial references.\n'
                         '- Avoid reprojections at all.\n'
                         '>>> help(Dataset) # for more informations.').format(an.messages)
                    raise ValueError(s)
            else:
                minx, maxx, miny, maxy = rect
                if minx != maxx and miny != maxy and not an.inverse_valid:
                    s = ('Error while checking if on-the-fly reprojection could be performed '
                         'between between Dataset\'s `sr_work` and vector\'s `sr_virtual`. '
                         'Cause: {} (The reprojection is too lossy).\n'
                         'Solutions:\n'
                         '- Pass those tests by reducing `buzz.env.significant`.\n'
                         '- Bypass those tests by passing `analyse_transformation=False` to the '
                         'Dataset\'s constructor.\n'
                         '- Use smaller coordinates.\n'
                         '- Use different spatial references.\n'
                         '- Avoid reprojections at all.\n'
                         '>>> help(Dataset) # for more informations.').format(an.messages)
                    raise ValueError(s)

        return to_work, to_virtual

    @staticmethod
    def _make_transfo(osr_transfo):
        """Wrap osr coordinate transformation input/output"""

        def _f(*args):
            nargs = len(args)

            if nargs == 1:
                # When coordinates in last dimension
                arr = np.asarray(args[0])
                assert arr.ndim >= 2
                ncoord = arr.shape[-1]
                assert 2 <= ncoord <= 3
                outshape = arr.shape

                arr = arr.reshape(int(arr.size / ncoord), ncoord)
                arr = osr_transfo(arr)
                arr = np.asarray(arr)
                arr = arr[:, 0:ncoord]
                arr = arr.reshape(outshape)
                return arr
            elif 2 <= nargs <= 3:
                # When coordinates in first dimension
                arr = np.asarray(args)
                assert arr.ndim == 2
                ncoord = nargs
                arr = np.moveaxis(arr, 0, 1)
                arr = osr_transfo(arr)
                arr = np.asarray(arr)
                arr = arr[:, 0:ncoord]
                arr = np.moveaxis(arr, 0, 1)
                return tuple(arr)
            else:
                assert False # pragma: no cover

        return _f

    def convert_footprint(self, fp, wkt):
        sr = osr.SpatialReference(wkt)
        _, to_virtual = self.get_transforms(sr, fp, 'work')
        if to_virtual:
            fp = fp.move(*to_virtual([fp.tl, fp.tr, fp.br]), round_coordinates=True)
        return fp

    @staticmethod
    def virtual_of_stored_given_mode(stored, work, fallback, forced):
        virtual = stored

        # Mode 4: If `ds` mode overrides file's stored
        if forced is not None:
            virtual = forced

        # Mode 3: If stored missing and `ds` provides a fallback
        if virtual is None and fallback is not None:
            virtual = fallback

        # Mode 2: If stored missing and `ds` does not provide a fallback
        if virtual is None and work is not None:
            raise ValueError("Missing source's spatial reference while using a `mode 2` Dataset")

        # Mode 1:
        if work is None:
            pass

        return virtual



import functools
import collections
import multiprocessing as mp
import multiprocessing.pool
import contextlib

import numpy as np

from buzzard._actors.message import Msg
from buzzard._actors.pool_job import ProductionJobWaiting, PoolJobWorking
from buzzard import _tools
from buzzard._gdal_file_raster import BackGDALFileRaster

class ActorReader(object):
    """Actor that takes care of reading cache tiles"""

    def __init__(self, raster):
        self._raster = raster
        self._back_ds = raster.back_ds
        self._alive = True
        io_pool = raster.io_pool
        if io_pool is not None:
            self._waiting_room_address = '/Pool{}/WaitingRoom'.format(id(io_pool))
            self._working_room_address = '/Pool{}/WorkingRoom'.format(id(io_pool))
            if isinstance(io_pool, mp.pool.ThreadPool):
                self._same_address_space = True
            elif isinstance(io_pool, mp.pool.Pool):
                self._same_address_space = False
            else: # pragma: no cover
                assert False, 'Type should be checked in facade'
        self._waiting_jobs = set()
        self._working_jobs = set()

        self._sample_array_per_prod_tile = (
            collections.defaultdict(dict)
        ) # type: Mapping[CachedQueryInfos, Mapping[int, np.ndarray]]
        self._missing_cache_fps_per_prod_tile = (
            collections.defaultdict(dict)
        ) # type: Mapping[CachedQueryInfos, Mapping[int, Set[Footprint]]]
        self.address = '/Raster{}/Reader'.format(self._raster.uid)

    @property
    def alive(self):
        return self._alive

    # ******************************************************************************************* **
    def receive_sample_cache_file_to_unique_array(self, qi, prod_idx, cache_fp, path):
        msgs = []

        if self._raster.io_pool is None:
            work = self._create_work_job(qi, prod_idx, cache_fp, path)
            work.func()
            msgs += self._commit_work_result(work, None)
        else:
            wait = Wait(self, qi, prod_idx, cache_fp, path)
            self._waiting_jobs.add(wait)
            msgs += [Msg(self._waiting_room_address, 'schedule_job', wait)]

        return msgs

    def receive_token_to_working_room(self, job, token):
        self._waiting_jobs.remove(job)
        work = self._create_work_job(job.qi, job.prod_idx, job.cache_fp, job.path)
        self._working_jobs.add(work)
        return [
            Msg(self._working_room_address, 'launch_job_with_token', work, token)
        ]

    def receive_job_done(self, job, result):
        self._working_jobs.remove(job)
        return self._commit_work_result(job, result)

    def receive_cancel_this_query(self, qi):
        msgs = []
        # Cancel waiting jobs
        jobs_to_kill = [
            job
            for job in self._waiting_jobs
            if job.qi == qi
        ]
        for job in jobs_to_kill:
            msgs += [Msg(self._waiting_room_address, 'unschedule_job', job)]
            self._waiting_jobs.remove(job)

        # Cancel working jobs
        jobs_to_kill = [
            job
            for job in self._working_jobs
            if job.qi == qi
        ]
        for job in jobs_to_kill:
            msgs += [Msg(self._working_room_address, 'cancel_job', job)]
            self._working_jobs.remove(job)

        # Clean datastructures
        if qi in self._sample_array_per_prod_tile:
            del self._sample_array_per_prod_tile[qi]
            del self._missing_cache_fps_per_prod_tile[qi]

        return msgs

    def receive_die(self):
        """Receive message: The raster was killed"""
        assert self._alive
        self._alive = False

        msgs = []
        for job in self._waiting_jobs:
            msgs += [Msg(self._waiting_room_address, 'unschedule_job', job)]
        for job in self._working_jobs:
            msgs += [Msg(self._working_room_address, 'cancel_job', job)]
        self._waiting_jobs.clear()
        self._working_jobs.clear()

        self._sample_array_per_prod_tile.clear()
        self._missing_cache_fps_per_prod_tile.clear()
        self._raster = None
        self._back_ds = None
        return msgs

    # ******************************************************************************************* **
    def _create_work_job(self, qi, prod_idx, cache_fp, path):
        if prod_idx not in self._sample_array_per_prod_tile[qi]:
            # Allocate sample array
            # If no interpolation or nodata conversion is necessary, this is the array that will be
            # returned in the output queue
            full_sample_fp = qi.prod[prod_idx].sample_fp
            self._sample_array_per_prod_tile[qi][prod_idx] = np.empty(
                np.r_[full_sample_fp.shape, len(qi.unique_channel_ids)],
                self._raster.dtype,
            )
            self._raster.debug_mngr.event(
                'object_allocated',
                self._sample_array_per_prod_tile[qi][prod_idx]
            )
            self._missing_cache_fps_per_prod_tile[qi][prod_idx] = set(qi.prod[prod_idx].cache_fps)

        dst_array = self._sample_array_per_prod_tile[qi][prod_idx]
        return Work(self, qi, prod_idx, cache_fp, path, dst_array)

    def _commit_work_result(self, job, result):
        if self._raster.io_pool is None or self._same_address_space:
            assert result is None
        else:
            job.dst_array_slice[:] = result

        dst_array = self._sample_array_per_prod_tile[job.qi][job.prod_idx]
        self._missing_cache_fps_per_prod_tile[job.qi][job.prod_idx].remove(job.cache_fp)

        # Perform fine grain garbage collection
        if len(self._missing_cache_fps_per_prod_tile[job.qi][job.prod_idx]) == 0:
            # Done reading for that `(qi, prod_idx)`
            del self._missing_cache_fps_per_prod_tile[job.qi][job.prod_idx]
            del self._sample_array_per_prod_tile[job.qi][job.prod_idx]

        if len(self._missing_cache_fps_per_prod_tile[job.qi]) == 0:
            # Not reading for that `qi`
            del self._missing_cache_fps_per_prod_tile[job.qi]
            del self._sample_array_per_prod_tile[job.qi]

        return [
            Msg('CacheExtractor', 'sampled_a_cache_file_to_the_array',
                job.qi, job.prod_idx, job.cache_fp, dst_array,
            )
        ]

    # ******************************************************************************************* **

class Wait(ProductionJobWaiting):
    def __init__(self, actor, qi, prod_idx, cache_fp, path):
        self.qi = qi
        self.prod_idx = prod_idx
        self.cache_fp = cache_fp
        self.sample_fp = cache_fp & qi.prod[prod_idx].sample_fp
        self.path = path
        super().__init__(actor.address, qi, prod_idx, 1, self.sample_fp)

class Work(PoolJobWorking):
    def __init__(self, actor, qi, prod_idx, cache_fp, path, dst_array):
        self.qi = qi
        self.prod_idx = prod_idx
        self.cache_fp = cache_fp
        raster = actor._raster
        full_sample_fp = qi.prod[prod_idx].sample_fp
        sample_fp = full_sample_fp & cache_fp

        dst_array_slice = dst_array[sample_fp.slice_in(full_sample_fp)]

        if actor._raster.io_pool is None or actor._same_address_space:
            func = functools.partial(
                _cache_file_read,
                path, cache_fp, actor._raster.dtype, qi.unique_channel_ids, sample_fp, dst_array_slice,
                actor._back_ds,
            )
        else:
            self.dst_array_slice = dst_array_slice
            func = functools.partial(
                _cache_file_read,
                path, cache_fp, actor._raster.dtype, qi.unique_channel_ids, sample_fp, None, None,
            )
        actor._raster.debug_mngr.event('object_allocated', func)
        super().__init__(actor.address, func)

def _cache_file_read(path, cache_fp, dtype, channel_ids, sample_fp, dst_opt, back_ds_opt):
    """
    Parameters
    ----------
    path: str
    cache_fp: Footprint
        Should be the Footprint of the cache file
    dtype: np.dtype
        Should be the dtype of the cache file
    channel_ids: sequence of int
    sample_fp: Footprint
        Rect of `cache_fp` to read
    dst_opt: None or np.ndarray
        optional destination for read
    """

    allocator = lambda: BackGDALFileRaster.open_file(path, 'GTiff', [], 'r')
    with contextlib.ExitStack() as stack:
        if back_ds_opt is None:
            gdal_ds = allocator()
        else:
            gdal_ds = stack.enter_context(back_ds_opt.acquire_driver_object(path, allocator))

        # Check raster
        if gdal_ds is None: # pragma: no cover
            raise RuntimeError("Could not open {path}, what happend to it?".format(
                path
            ))
        if (gdal_ds.RasterXSize, gdal_ds.RasterYSize) != tuple(cache_fp.rsize): # pragma: no cover
            raise RuntimeError('{} was expected to have rsize {}, not {}'.format(
                path,
                tuple(cache_fp.rsize),
                (gdal_ds.RasterXSize, gdal_ds.RasterYSize),
            ))
        stored_dtype = _tools.conv.dtype_of_gdt_downcast(gdal_ds.GetRasterBand(1).DataType)
        if dtype != stored_dtype: # pragma: no cover
            raise RuntimeError('{} was expected to have dtype {}, not {}'.format(
                path,
                dtype,
                stored_dtype,
            ))

        # Allocate if ProcessPool
        if dst_opt is None:
            dst = np.empty(np.r_[sample_fp.shape, len(channel_ids)], dtype)
            ret = dst
        else:
            dst = dst_opt
            ret = None

        # Perform read
        rtlx, rtly = cache_fp.spatial_to_raster(sample_fp.tl)
        for i, ci in enumerate(channel_ids):
            b = gdal_ds.GetRasterBand(ci + 1)
            a = b.ReadAsArray(
                int(rtlx),
                int(rtly),
                int(sample_fp.rsizex),
                int(sample_fp.rsizey),
                buf_obj=dst[..., i],
            )
            del b
            if a is None: # pragma: no cover
                raise RuntimeError('Could not read channel_id {}'.format(ci))
    del gdal_ds

    # Return
    return ret



# pylint: disable=redefined-outer-name

from __future__ import division, print_function
import itertools

import numpy as np
import pytest
import shapely.geometry as sg

import buzzard as buzz
from buzzard.test.tools import fpeq
from buzzard.test import make_tile_set

LETTERS = "ABCDEFGHI"


@pytest.fixture(scope='module')
def fps():
    """
    See make_tile_set
    A B C
    D E F
    G H I
    """
    return make_tile_set.make_tile_set(3, [0.1, -0.1])

@pytest.fixture(scope='module')
def fps1px():
    """
    See make_tile_set
    A B C
    D E F
    G H I
    """
    return make_tile_set.make_tile_set(3, [1, -1], (1, -1))


def test_rectangles(fps):
    for a, b in itertools.combinations_with_replacement(fps.values(), 2):
        if a.poly.overlaps(b.poly) or a.poly.covers(b.poly) or b.poly.covers(a.poly):
            assert fpeq(
                a.intersection(b),
                b.intersection(a)
            )
    dfs = [
        fps.DF,
        fps.AF.intersection(fps.DI),
        fps.DF.intersection(fps.AI),
        fps.DF.intersection(fps.AF),
        fps.DF.intersection(fps.DI),
    ]
    for a in dfs:
        assert fpeq(a, dfs[0])
    bhs = [
        fps.BH,
        fps.AH.intersection(fps.BI),
        fps.BH.intersection(fps.AI),
        fps.BH.intersection(fps.AH),
        fps.BH.intersection(fps.BI),
    ]
    for a in bhs:
        assert fpeq(a, bhs[0])

    assert fpeq(
        fps.E,
        fps.E.intersection(fps.E),
        fps.AI.intersection(
            fps.BH, fps.DF,
            fps.AH, fps.BI,
            fps.AF, fps.DI,
            fps.AI,
        ),
    )

def test_points(fps1px):
    """Test points at fps.E's points of interest"""
    fps = fps1px
    assert fpeq(
        fps.E,
        fps.AI.intersection(sg.Point(*fps.E.c)),
        fps.AI.intersection(sg.Point(*fps.E.t)),
        fps.AI.intersection(sg.Point(*fps.E.l)),
        fps.AI.intersection(sg.Point(*fps.E.tl)),
    )
    assert fpeq(
        fps.I,
        fps.AI.intersection(sg.Point(*fps.E.br)),
    )
    assert fpeq(
        fps.H,
        fps.AI.intersection(sg.Point(*fps.E.bl)),
        fps.AI.intersection(sg.Point(*fps.E.b)),
    )
    assert fpeq(
        fps.F,
        fps.AI.intersection(sg.Point(*fps.E.tr)),
        fps.AI.intersection(sg.Point(*fps.E.r)),
    )

def test_lines(fps1px):
    """Test small lines centered at fps.E's points of interest
    Test all 3 orientations
    """
    fps = fps1px

    def _f(coords, axes):
        """create small line around point"""
        axes = np.asarray(axes)
        assert 1 <= axes.sum() <= 2
        epsilon = 10 ** -(buzz.env.significant - 1)
        tl = coords - (epsilon, epsilon) * axes * (1, -1)
        br = coords + (epsilon, epsilon) * axes * (1, -1)
        return sg.LineString([tl, br])

    # Middle, size 1
    assert fpeq(
        fps.E,
        fps.AI.intersection(_f(fps.E.c, (1, 0))),
        fps.AI.intersection(_f(fps.E.c, (1, 1))),
        fps.AI.intersection(_f(fps.E.c, (0, 1))),
        fps.AI.intersection(_f(fps.E.l, (0, 1))),
        fps.AI.intersection(_f(fps.E.t, (1, 0))),
    )

    # Connectivity 4, size 1&2
    assert fpeq(
        fps.BE,
        fps.AI.intersection(_f(fps.E.t, (1, 1))),
        fps.AI.intersection(_f(fps.E.t, (0, 1))),
        fps.AI.intersection(_f(fps.E.tl, (0, 1))),
    )

    assert fpeq(
        fps.EF,
        fps.AI.intersection(_f(fps.E.r, (1, 0))),
        fps.AI.intersection(_f(fps.E.r, (1, 1))),
        fps.AI.intersection(_f(fps.E.tr, (1, 0))),
    )
    assert fpeq(
        fps.F,
        fps.AI.intersection(_f(fps.E.r, (0, 1))),
    )

    assert fpeq(
        fps.EH,
        fps.AI.intersection(_f(fps.E.b, (1, 1))),
        fps.AI.intersection(_f(fps.E.b, (0, 1))),
        fps.AI.intersection(_f(fps.E.bl, (0, 1))),
    )
    assert fpeq(
        fps.H,
        fps.AI.intersection(_f(fps.E.b, (1, 0))),
    )

    assert fpeq(
        fps.DE,
        fps.AI.intersection(_f(fps.E.l, (1, 0))),
        fps.AI.intersection(_f(fps.E.l, (1, 1))),
        fps.AI.intersection(_f(fps.E.tl, (1, 0))),
    )

    # Connectivity 8, size 4
    assert fpeq(
        fps.AE,
        fps.AI.intersection(_f(fps.E.tl, (1, 1))),
    )
    assert fpeq(
        fps.BF,
        fps.AI.intersection(_f(fps.E.tr, (1, 1))),
    )
    assert fpeq(
        fps.EI,
        fps.AI.intersection(_f(fps.E.br, (1, 1))),
    )
    assert fpeq(
        fps.DH,
        fps.AI.intersection(_f(fps.E.bl, (1, 1))),
    )

    # Connectivity 8, size 2
    assert fpeq(
        fps.CF,
        fps.AI.intersection(_f(fps.E.tr, (0, 1))),
    )
    assert fpeq(
        fps.FI,
        fps.AI.intersection(_f(fps.E.br, (0, 1))),
    )
    assert fpeq(
        fps.HI,
        fps.AI.intersection(_f(fps.E.br, (1, 0))),
    )
    assert fpeq(
        fps.GH,
        fps.AI.intersection(_f(fps.E.bl, (1, 0))),
    )

    # Bonus (diagonals)
    assert fpeq(
        fps.AI,
        fps.AI.intersection(sg.LineString([fps.AI.tl, fps.AI.br])),
        fps.AI.intersection(sg.LineString([fps.AI.bl, fps.AI.tr])),
        fps.AI.intersection(sg.LineString([fps.A.t, fps.I.b])),
        fps.AI.intersection(sg.LineString([fps.A.l, fps.I.r])),
        fps.AI.intersection(sg.LineString([fps.A.c, fps.I.c])),
    )

    # Bonus 2 (multi point line)
    assert fpeq(
        fps.AI,
        fps.AI.intersection(sg.LineString([
            fps.A.c, fps.F.c, fps.G.c,
        ])),
        fps.AI.intersection(sg.LineString([
            fps.A.c, fps.C.c, fps.G.r,
        ])),
    )


class _FtPoly(object):

    def __init__(self, data):
        self.__geo_interface__ = data


def test_corner_cases(fps1px):
    fps = fps1px

    with pytest.raises(ValueError):
        fps.A.intersection()
    with pytest.raises(ValueError):
        fps.A.intersection(fps.A, hello=True)
    with pytest.raises(TypeError):
        fps.A.intersection(42)
    assert fpeq(
        fps.BH,
        fps.AH.intersection(_FtPoly(fps.BI.__geo_interface__)),
    )
    with pytest.raises(ValueError):
        fps.A.intersection(fps.A, scale='hello')
    with pytest.raises(ValueError):
        fps.A.intersection(fps.A, rotation='hello')
    with pytest.raises(ValueError):
        fps.A.intersection(fps.A, alignment='hello')

    # reso
    assert fpeq(
        buzz.Footprint(rsize=[2, 6], size=fps.BH.size, tl=fps.BH.tl),
        fps.AH.intersection(fps.BI, scale=0.5),
        fps.AH.intersection(fps.BI, scale=[0.5, -0.5]),
        fps.AH.intersection(fps.BI, scale=[0.5]),
    )
    with pytest.raises(ValueError):
        fps.AH.intersection(fps.BI, scale=[])
    with pytest.raises(ValueError):
        fps.AH.intersection(fps.BI, scale=0)

    lowest = fps.BH.intersection(fps.BH, scale=0.5)
    highest = fps.BH.intersection(fps.BH, scale=1.0)
    assert fpeq(
        lowest,
        lowest.intersection(highest, scale='lowest'),
        highest.intersection(lowest, scale='lowest'),
    )
    assert fpeq(
        highest,
        highest.intersection(lowest, scale='highest'),
        lowest.intersection(highest, scale='highest'),
    )

    # rot / align
    assert fpeq(
        fps.BH,
        fps.AH.intersection(fps.BI, rotation=0),
        fps.AH.intersection(fps.BI, alignment=fps.BH.tl),
    )
    with pytest.raises(ValueError):
        fps.AH.intersection(fps.BI, alignment=[])
    assert fpeq(
        buzz.Footprint(
            rsize=[1 + 1, 3 + 1], size=fps.BH.size * [2/1, 4/3], tl=fps.BH.tl - [0.5, -0.5]
        ),
        fps.BH.intersection(fps.BH, alignment=[0.5, 0.5]),
    )
    assert fpeq(
        fps.BH,
        fps.BH.intersection(fps.BH, alignment='tl'),
    )
    with buzz.Env(allow_complex_footprint=True):
        for angle in np.r_[0:180:13j]:
            rotated = fps.E.intersection(fps.E, rotation=angle)
            nofit = angle % 90 != 0
            if nofit:
                assert tuple(rotated.rsize) == (2, 2)
            else:
                assert tuple(rotated.rsize) == (1, 1)
            assert all(np.around(rotated.scale, 3) == (1, -1))
            assert np.around(rotated.angle, 3) == angle
            diff = rotated.poly - fps.E.poly
            if nofit:
                assert np.around(diff.area, 3) == 3.0
            else:
                assert np.around(diff.area, 3) == 0.0
            dot = np.dot(fps.E.lrvec / fps.E.w, rotated.lrvec / rotated.w)
            angle_real = np.arccos(dot) / np.pi * 180
            assert np.around(angle_real) == angle

    # homo
    assert fpeq(
        fps.BH,
        fps.BH.intersection(fps.BH, homogeneous=True),
    )
    with pytest.raises(ValueError, match='grid'):
        fps.BH.intersection(
            fps.BH.intersection(fps.BH, scale=0.5),
            homogeneous=True,
        )
    with pytest.raises(ValueError, match='grid'):
        fps.BH.intersection(
            fps.BH.intersection(fps.BH, alignment=[0.5, 0.5]),
            homogeneous=True,
        )
    with buzz.Env(allow_complex_footprint=True):
        with pytest.raises(ValueError, match='grid'):
            fps.AH.intersection(
                fps.E.intersection(fps.E, rotation=42),
                homogeneous=True,
            )

    # fit
    stripe = sg.Polygon([fps.A.tr, fps.I.tr, fps.I.bl, fps.A.bl])
    assert fpeq(
        fps.AI,
        fps.AI.intersection(stripe),
        fps.AI.intersection(stripe, rotation=0),
    )
    with buzz.Env(allow_complex_footprint=True):
        assert fpeq(
            fps.AI.intersection(stripe, rotation='fit'),
            fps.AI.intersection(stripe, rotation=45),
        )

    # misc
    with pytest.raises(ValueError, match='touch'):
        fps.A.intersection(fps.B)
    with pytest.raises(ValueError, match='empty'):
        fps.A.intersection(fps.C)
    with pytest.raises(ValueError, match='touch'):
        fps.A.intersection(fps.D)



"""Resampling tests for *Raster.get_data() methods (not testing rotations)

TODO: Unit test set_data with resampling
TODO: Unit test `remap` mixin on one side and `getsetdata_remap` elsewhere
"""

# pylint: disable=redefined-outer-name

from __future__ import division, print_function
import itertools
import os
import uuid
import tempfile

import numpy as np
import pytest

from buzzard import Footprint, Dataset

# CONSTANTS - INTERPOLATIONS ******************************************************************** **
INTERPOLATIONS = [
    'cv_area',
    'cv_linear',
    'cv_nearest',
    'cv_cubic',
    'cv_lanczos4',
]
INTERPOLATIONS_MAX_VALUE_ERROR = {
    'cv_area': 0.023,
    'cv_linear': 0.023,
    'cv_nearest': 1.,
    'cv_cubic': 0.11,
    'cv_lanczos4': 0.23,
}
INTERPOLATIONS_INSIDE_UNCERTAINTY_BORDER = {
    'cv_area': 0,
    'cv_linear': 0,
    'cv_nearest': 0,
    'cv_cubic': 1, # Massive spikes near borders. Nodata holes near topleft borders.
    'cv_lanczos4': 3, # Nodata holes near topleft borders.
}
INTERPOLATIONS_OUTSIDE_UNCERTAINTY_BORDER = {
    'cv_area': 1,
    'cv_linear': 1,
    'cv_nearest': 2,
    'cv_cubic': 2,
    'cv_lanczos4': 2,
}

# CONSTANTS - TIF GENERATION ******************************************************************** **
TIF_NODATA = 200
TIF_FP = Footprint(
    tl=(100, 110), size=(10, 10), rsize=(10, 10)
)
TIF_VALUES = np.add(*TIF_FP.meshgrid_raster).astype('float32')
TIF_VALUES[:1] = TIF_NODATA
TIF_VALUES[:, :2] = TIF_NODATA

XS, YS = TIF_FP.meshgrid_spatial
TIF_DATAMASK = TIF_VALUES != TIF_NODATA
TIF_DATA_MAXY = YS[TIF_DATAMASK].max()
TIF_DATA_MINY = YS[TIF_DATAMASK].min()
TIF_DATA_MAXX = XS[TIF_DATAMASK].max()
TIF_DATA_MINX = XS[TIF_DATAMASK].min()
TIF_MAXY = YS.max()
TIF_MINY = YS.min()
TIF_MAXX = XS.max()
TIF_MINX = XS.min()
del XS, YS

# CONSTANTS - SCENARIOS ************************************************************************* **
# All tested sizes should be multiples of all resolutions tested ******************************** **
DATA_FULL_LOADING = {
    'size': [12],
    'scale': [3/1, 3/2, 3/3, 3/4, 3/6, 3/7],
    'offset': [0., 0.5, np.sqrt(2), 2., np.sqrt(2) + 1],
}

ZONES_EDGES_TESTS = {
    'size': [1, 9, 10, 14],
    'scale': [1.],
    'offset': [-15, -4, -0.5, 0., 1.5, 2., 2.5],
}

SCENARIOS = [DATA_FULL_LOADING, ZONES_EDGES_TESTS]

# FIXTURES ************************************************************************************** **
@pytest.fixture(scope='module')
def ds():
    return Dataset(allow_interpolation=1)

@pytest.fixture(
    scope='module',
    params=[
        ('GTiff', 1, 'float32', TIF_NODATA),
        ('GTiff', 3, 'uint8', None),
        ('MEM', 1, 'float32', TIF_NODATA),
        ('MEM', 3, 'uint8', None),
        ('numpy', 1, 'float32', None),
        ('numpy', 3, 'uint8', TIF_NODATA),
    ],
)
def rast(request, ds):
    """Fixture for the dataset creation"""
    fp = TIF_FP
    driver, channel_count, dtype, nodata = request.param
    if driver == 'numpy':
        rast = ds.awrap_numpy_raster(
            fp,
            np.dstack([TIF_VALUES.copy().astype(dtype=dtype)] * channel_count),
            channels_schema=dict(nodata=nodata),
            sr=None,
            mode='r',
        )
    elif driver == 'MEM':
        rast = ds.acreate_raster(
            '', fp, dtype, channel_count, channels_schema=dict(nodata=nodata), driver='MEM',
        )
        for band_id in range(1, len(rast) + 1):
            rast.set_data(TIF_VALUES, band=band_id)
    else:
        path = '{}/{}.tif'.format(tempfile.gettempdir(), uuid.uuid4())
        rast = ds.acreate_raster(
            path, fp, dtype, channel_count, channels_schema=dict(nodata=nodata), driver=driver
        )
        for band_id in range(1, len(rast) + 1):
            rast.set_data(TIF_VALUES, band=band_id)
    yield rast
    if driver in {'numpy', 'MEM'}:
        rast.close()
    else:
        rast.delete()

def pytest_generate_tests(metafunc):
    values = []
    for scenar in SCENARIOS:
        it = itertools.product(
            scenar['size'], scenar['scale'], scenar['offset'], [[1, 1], [-1, 1], [1, -1], [-1, -1]], INTERPOLATIONS,
        )
        values += it
    metafunc.parametrize(
        argnames='size1,scale1,offset1,offset_factor2,interpolation',
        argvalues=values
    )

@pytest.fixture()
def fp(size1, scale1, offset1, offset_factor2):
    tl = TIF_FP.tl + [offset1, offset1] * np.asarray(offset_factor2)
    size = [size1, size1]
    rsize = (size / np.asarray(scale1)).astype(int)
    assert (size == rsize * scale1).all()
    fp = Footprint(tl=tl, size=size, rsize=np.abs(rsize))
    return fp

# TESTS ***************************************************************************************** **
def test_getdata(rast, fp, interpolation):
    if rast.nodata is None:
        all_res = rast.get_data(band=-1, fp=fp, interpolation=interpolation, dst_nodata=TIF_NODATA)
    else:
        all_res = rast.get_data(band=-1, fp=fp, interpolation=interpolation)

    all_res = np.atleast_3d(all_res)

    for i in range(len(rast)):
        res = all_res[..., i]

        # 1 - Assert absent nodata within input pixels (pixels being points not areas)
        xs, ys = fp.meshgrid_spatial
        b = INTERPOLATIONS_INSIDE_UNCERTAINTY_BORDER[interpolation]
        deep_inside_data_mask = (
            (xs > TIF_DATA_MINX + TIF_FP.pxsizex * b) &
            (xs < TIF_DATA_MAXX - TIF_FP.pxsizex * b) &
            (ys > TIF_DATA_MINX + TIF_FP.pxsizex * b) &
            (ys < TIF_DATA_MAXX - TIF_FP.pxsizex * b)
        )
        assert np.all(
            res[deep_inside_data_mask] != TIF_NODATA
        )

        # 2 - Test the difference of values between two neighboring columns/rows
        if deep_inside_data_mask.any():
            xs, ys = fp.meshgrid_raster
            data_slice = (
                slice(ys[deep_inside_data_mask].min(), ys[deep_inside_data_mask].max() + 1),
                slice(xs[deep_inside_data_mask].min(), xs[deep_inside_data_mask].max() + 1),
            )
            below_minus_above = np.diff(res[data_slice], axis=0)
            right_minus_left = np.diff(res[data_slice], axis=1)

            vertical_errors = np.abs(below_minus_above - fp.pxsizex)
            horizontal_errors = np.abs(right_minus_left - fp.pxsizex)
            if np.issubdtype(rast.dtype, np.integer):
                maxerr = np.ceil(INTERPOLATIONS_MAX_VALUE_ERROR[interpolation])
            else:
                maxerr = INTERPOLATIONS_MAX_VALUE_ERROR[interpolation]
            if vertical_errors.size:
                assert vertical_errors.max() <= maxerr
            if horizontal_errors.size:
                assert horizontal_errors.max() <= maxerr

        # 3 - Assert only nodata far from data
        xs, ys = fp.meshgrid_spatial
        b = INTERPOLATIONS_OUTSIDE_UNCERTAINTY_BORDER[interpolation]
        if rast.nodata is not None:
            far_outside_data_mask = (
                (xs < TIF_DATA_MINX - TIF_FP.pxsizex * b) |
                (xs > TIF_DATA_MAXX + TIF_FP.pxsizex * b) |
                (ys < TIF_DATA_MINX - TIF_FP.pxsizex * b) |
                (ys > TIF_DATA_MAXX + TIF_FP.pxsizex * b)
            )
        else:
            far_outside_data_mask = (
                (xs < TIF_MINX - TIF_FP.pxsizex * b) |
                (xs > TIF_MAXX + TIF_FP.pxsizex * b) |
                (ys < TIF_MINX - TIF_FP.pxsizex * b) |
                (ys > TIF_MAXX + TIF_FP.pxsizex * b)
            )
        assert np.all(
            res[far_outside_data_mask] == TIF_NODATA
        )



import time
from concurrent.futures import ThreadPoolExecutor

import buzzard as buzz
import numpy as np

def work(i):
    j = i % 9 + 1
    assert buzz.env.significant == 42
    with buzz.Env(significant=j):
        assert buzz.env.significant == j
        time.sleep(np.random.rand() / 100)
        assert buzz.env.significant == j
    assert buzz.env.significant == 42

def test_thread_pool():
    with buzz.Env(significant=42):
        with ThreadPoolExecutor(10) as ex:
            it = ex.map(
                work,
                range(100),
            )
            list(it)



import numpy as np

def concat_arrays(fp, array_per_fp, _):
    """Concatenate arrays from `array_per_fp` to form `fp`.

    This function is meant to be fed to the `merge_arrays` parameter when constructing a recipe.
    """
    # Allocate
    for a in array_per_fp.values():
        channel_count = a.shape[-1]
        dtype = a.dtype
        break
    arr = np.empty(np.r_[fp.shape, channel_count], dtype)

    # Burn
    for tile, tile_arr in array_per_fp.items():
        assert tuple(tile.shape) == tile_arr.shape[:2]
        slices = tile.slice_in(fp)
        arr[slices] = tile_arr

    # Return
    return arr



import os
import sys

from setuptools import setup, find_packages

# Requirements
# https://caremad.io/posts/2013/07/setup-vs-requirement/
reqs = [
    'opencv-python>=3.1.0',
    'gdal>=2.3.3',
    'shapely>=1.5.16',
    'affine>=2.0.0',
    'scipy>=0.19.1',
    'pint>=0.8.1',
    'six>=1.11.0',
    'Rtree>=0.8.3',
    'sortedcontainers>=1.5.9',
]

if (3, 4) <= sys.version_info < (3, 5):
    # https://github.com/scikit-image/scikit-image/issues/3820
    reqs += ['scikit-image<=0.14.1', 'numpy<1.16']
else:
    reqs += ['scikit-image>=0.14.0', 'numpy>=1.15']

if sys.version_info < (3, 2):
    reqs += ['chainmap>=1.0.2']

readme_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'README.md'
)
readme = open(readme_path, 'rb').read().decode("UTF-8")

# Classifiers
# https://pypi.python.org/pypi?%3Aaction=list_classifiers
classifiers = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Intended Audience :: Information Technology',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Topic :: Scientific/Engineering :: GIS',
]

setup(
    name='buzzard',
    version='0.6.5',
    author='ngoguey',
    author_email='ngoguey@student.42.fr',
    license='Apache License 2.0',
    maintainer='ngoguey',
    maintainer_email='ngoguey@student.42.fr',
    description='GIS files manipulations',
    long_description=readme,
    long_description_content_type='text/markdown',
    classifiers=classifiers,
    url='https://github.com/earthcube-lab/buzzard',
    download_url='https://github.com/earthcube-lab/buzzard/archive/0.6.5.tar.gz',
    keywords=['gdal gis raster shp dxf tif vector'],
    packages=find_packages(),
    install_requires=reqs,
    python_requires='>=3.4',
)



"""
IMPORTANT: When you import :code:`buzzard` for the first time, you should always do it from the main thread.
"""

__version__ = "0.6.5"

# Import osgeo before cv2
import osgeo as _
import cv2 as _

# Public classes
from buzzard._footprint import Footprint
from buzzard._dataset import (
    Dataset,
    open_raster,
    open_vector,
    create_raster,
    create_vector,
    wrap_numpy_raster
)
from buzzard._dataset import (
    DataSource, # Deprecated
)

from buzzard._env import Env

# Source's abstract classes
# Public methods, but always instanciated by Dataset, never by user.
from buzzard._a_source import ASource
from buzzard._a_source_raster import ASourceRaster
from buzzard._a_source_vector import ASourceVector

from buzzard._a_stored import AStored
from buzzard._a_stored_raster import AStoredRaster
from buzzard._a_stored_vector import AStoredVector

from buzzard._a_emissary import AEmissary
from buzzard._a_emissary_raster import AEmissaryRaster
from buzzard._a_emissary_vector import AEmissaryVector

from buzzard._a_pooled_emissary import APooledEmissary
from buzzard._a_pooled_emissary_raster import APooledEmissaryRaster
from buzzard._a_pooled_emissary_vector import APooledEmissaryVector

from buzzard._a_async_raster import AAsyncRaster
from buzzard._a_raster_recipe import ARasterRecipe

# Source's concrete classes
# Public methods, but always instanciated by Dataset, never by user.
from buzzard._gdal_file_raster import GDALFileRaster
from buzzard._gdal_mem_raster import GDALMemRaster
from buzzard._numpy_raster import NumpyRaster

from buzzard._gdal_file_vector import GDALFileVector
from buzzard._gdal_memory_vector import GDALMemoryVector

from buzzard._cached_raster_recipe import CachedRasterRecipe

# Misc classes
# Public methods, but always instanciated by Dataset, never by user.
from buzzard._dataset_pools_container import PoolsContainer

# Misc
from buzzard._env import env

# Public submodules
import buzzard.utils
import buzzard.srs
import buzzard.algo



""">>> help(buzz.Dataset)"""

# pylint: disable=too-many-lines
import sys
import pathlib
import itertools
from types import MappingProxyType
import os

from osgeo import osr
import numpy as np

from buzzard._tools import conv, deprecation_pool
from buzzard._tools import GDALErrorCatcher as Catch

from buzzard._footprint import Footprint
from buzzard import _tools
from buzzard._dataset_back import BackDataset
from buzzard._a_source import ASource
from buzzard._gdal_file_raster import GDALFileRaster, BackGDALFileRaster
from buzzard._gdal_file_vector import GDALFileVector, BackGDALFileVector
from buzzard._gdal_mem_raster import GDALMemRaster
from buzzard._gdal_memory_vector import GDALMemoryVector
from buzzard._dataset_register import DatasetRegisterMixin
from buzzard._numpy_raster import NumpyRaster
from buzzard._cached_raster_recipe import CachedRasterRecipe
from buzzard._a_pooled_emissary import APooledEmissary
import buzzard.utils

class Dataset(DatasetRegisterMixin):
    """**Dataset** is a class that stores references to sources. A source is either a raster, or a
    vector. A `Dataset` allows:

    + quick manipulations by optionally assigning a key to each registered source, \
      (see :ref:`Sources Registering` below)
    + closing all source at once by closing the Dataset object.

    But also inter-sources operations, like:

    + spatial reference harmonization (see :ref:`On the fly re-projections in buzzard` below),
    + workload scheduling on pools when using async rasters (see :ref:`Scheduler` below),
    + other features in the future (like data visualization).

    For actions specific to opened sources, see those classes:

    - :doc:`source_gdal_file_raster`
    - :doc:`source_gdal_mem_raster`
    - :doc:`source_numpy_raster`
    - :doc:`source_cached_raster_recipe`
    - :doc:`source_gdal_file_vector`
    - :doc:`source_gdal_memory_vector`

    .. warning::
        This class is not equivalent to the `gdal.Dataset` class.

    Parameters
    ----------
    sr_work: None or string
        In order to set a spatial reference, use a string that can be `converted to WKT by GDAL
        <https://gdal.org/doxygen/classOGRSpatialReference.html#aec3c6a49533fe457ddc763d699ff8796>`_.

        (see :ref:`On the fly re-projections in buzzard` below)
    sr_fallback: None or string
        In order to set a spatial reference, use a string that can be `converted to WKT by GDAL
        <https://gdal.org/doxygen/classOGRSpatialReference.html#aec3c6a49533fe457ddc763d699ff8796>`_.

        (see :ref:`On the fly re-projections in buzzard` below)
    sr_forced: None or string
        In order to set a spatial reference, use a string that can be `converted to WKT by GDAL
        <https://gdal.org/doxygen/classOGRSpatialReference.html#aec3c6a49533fe457ddc763d699ff8796>`_.

        (see :ref:`On the fly re-projections in buzzard` below)
    analyse_transformation: bool
        Whether or not to perform a basic analysis on two `sr` to check their compatibility.

        if True: Read the `buzz.env.significant` variable and raise an exception if a spatial
        reference conversions is too lossy in precision.

        if False: Skip all checks.

        (see :ref:`On the fly re-projections in buzzard` below)
    allow_none_geometry: bool
        Whether or not a vector geometry should raise an exception when encountering a None geometry
    allow_interpolation: bool
        Whether or not a raster geometry should raise an exception when remapping with interpolation
        is necessary.
    max_active: nbr >= 1
        Maximum number of pooled sources active at the same time.
        (see :ref:`Sources activation / deactivation` below)
    debug_observers: sequence of object
        Entry points to observe what is happening in the Dataset's sheduler.

    Examples
    --------
    >>> import buzzard as buzz

    Creating a Dataset.

    >>> ds = buzz.Dataset()

    Opening a file and registering it under the 'roofs' key. There are four ways to the access an
    opened source.

    >>> r = ds.open_vector('roofs', 'path/to/roofs.shp')
    ... feature_count = len(ds.roofs)
    ... feature_count = len(ds['roofs'])
    ... feature_count = len(ds.get('roofs'))
    ... feature_count = len(r)

    Opening a file anonymously. There is only one way to access the source.

    >>> r = ds.aopen_raster('path/to/dem.tif')
    ... data_type = r.dtype

    Opening, reading and closing two raster files with context management.

    >>> with ds.open_raster('rgb', 'path/to/rgb.tif').close:
    ...     data_type = ds.rgb.fp
    ...     arr = ds.rgb.get_data()

    >>> with ds.aopen_raster('path/to/rgb.tif').close as rgb:
    ...     data_type = rgb.dtype
    ...     arr = rgb.get_data()

    Creating two files

    >>> ds.create_vector('targets', 'path/to/targets.geojson', 'point', driver='GeoJSON')
    ... geometry_type = ds.targets.type

    >>> with ds.acreate_raster('/tmp/cache.tif', ds.dem.fp, 'float32', 1).delete as cache:
    ...     file_footprint = cache.fp
    ...     cache.set_data(dem.get_data())

    Sources Types
    -------------
    - Raster sources
        - GDAL drivers http://www.gdal.org/formats_list.html (e.g. 'GTIff', 'JPEG', 'PNG', ...)
        - numpy.ndarray
        - recipes
    - Vector sources
        - OGR drivers: https://www.gdal.org/ogr_formats.html (e.g. 'ESRI Shapefile', 'GeoJSON', 'DXF', ...)

    .. _Sources Registering:
    Sources Registering
    -------------------
    There are always two ways to create a source, with a key or anonymously.

    When creating a source using a key, said key (e.g. the string "my_source_name") must be provided
    by user. Each key identify one source and should thus be unique. There are then three ways to
    access that source:

    - from the object returned by the method that created the source,
    - from the Dataset with the attribute syntax: `ds.my_source_name`,
    - from the Dataset with the item syntax: ds["my_source_name"].

    All keys should be unique.

    When creating a source anonymously you don't have to provide a key, but the only way to access
    this source is to use the object returned by the method that created the source.

    .. _Sources activation / deactivation:
    Sources activation / deactivation
    ---------------------------------

    The sources that inherit from `APooledEmissary` (like `GDALFileVector` and `GDALFileRaster`) are
    flexible about their underlying driver object. Those sources may be temporary deactivated
    (useful to limit the number of file descriptors active), or activated multiple time at the
    same time (useful to perfom concurrent reads).

    Those sources are automatically activated and deactivated given the current needs and
    constraints. Setting a `max_active` lower than `np.inf` in the Dataset constructor, will
    ensure that no more than `max_active` driver objects are active at the same time, by
    deactivating the LRU ones.

    .. _On the fly re-projections in buzzard:
    On the fly re-projections in buzzard
    ------------------------------------

    A Dataset may perform spatial reference conversions on the fly, like a GIS does. Several
    modes are available, a set of rules define how each mode work. Those conversions concern both
    read operations and write operations, all are performed by the OSR library.

    Those conversions are only perfomed on vector's data/metadata and raster's Footprints.
    This implies that classic raster warping is not included (yet) in those conversions, only raster
    shifting/scaling/rotation work.

    The `z` coordinates of vectors geometries are also converted, on the other hand elevations are
    not converted in DEM rasters.

    If `analyse_transformation` is set to `True` (default), all coordinates conversions are
    tested against `buzz.env.significant` on file opening to ensure their feasibility or
    raise an exception otherwise. This system is naive and very restrictive, use with caution.
    Although, disabling those tests is not recommended, ignoring floating point precision errors
    can create unpredictable behaviors at the pixel level deep in your code. Those bugs can be
    witnessed when zooming to infinity with tools like `qgis` or `matplotlib`.

    On the fly re-projections in buzzard - Terminology
    --------------------------------------------------
    `sr`
      Spatial reference
    `sr_work`
      The sr of all interactions with a Dataset (i.e. Footprints, extents, Polygons...),
      may be None.
    `sr_stored`
      The sr that can be found in the metadata of a raster/vector storage, may be None.
    `sr_virtual`
      The sr considered to be written in the metadata of a raster/vector storage, it is
      often the same as `sr_stored`. When a raster/vector is read, a conversion is performed from
      `sr_virtual` to `sr_work`. When writing vector data, a conversion is performed from
      `sr_work` to `sr_virtual`.
    `sr_forced`
      A `sr_virtual` provided by user to ignore all `sr_stored`. This is for example
      useful when the `sr` stored in the input files are corrupted.
    `sr_fallback`
      A `sr_virtual` provided by user to be used when `sr_stored` is missing. This is
      for example useful when an input file can't store a `sr` (e.g. DFX).

    On the fly re-projections in buzzard -  Dataset parameters and modes
    --------------------------------------------------------------------

    +------+----------+--------------+------------+-------------------------------------------------------------------------------+
    | mode | sr_work  | sr_fallback  | sr_forced  | How is the `sr_virtual` of a source determined                                |
    +======+==========+==============+============+===============================================================================+
    | 1    | None     | None         | None       | Use `sr_stored`, no conversion is performed for the lifetime of this Dataset  |
    +------+----------+--------------+------------+-------------------------------------------------------------------------------+
    | 2    | string   | None         | None       | Use `sr_stored`, if None raises an exception                                  |
    +------+----------+--------------+------------+-------------------------------------------------------------------------------+
    | 3    | string   | string       | None       | Use `sr_stored`, if None it is considered to be `sr_fallback`                 |
    +------+----------+--------------+------------+-------------------------------------------------------------------------------+
    | 4    | string   | None         | string     | Use `sr_forced`                                                               |
    +------+----------+--------------+------------+-------------------------------------------------------------------------------+

    On the fly re-projections in buzzard - Use cases
    ------------------------------------------------
    - If all opened files are known to be written in a same sr in advance, use `mode 1`.
        No conversions will be performed, this is the safest way to work.

    - If all opened files are known to be written in the same sr but you wish to work in a \
    different sr, use `mode 4`.
        The huge benefit of this mode is that the `driver` specific behaviors
        concerning spatial references have no impacts on the data you manipulate.

    - On the other hand if you don't have a priori information on files' `sr`, `mode 2` or \
    `mode 3` should be used.
         .. warning::
             Side note: Since the GeoJSON driver cannot store a `sr`, it is impossible to open or
             create a GeoJSON file in `mode 2`.

    On the fly re-projections in buzzard - Examples
    -----------------------------------------------
    mode 1 - No conversions at all

    >>> ds = buzz.Dataset()

    mode 2 - Working with WGS84 coordinates

    >>> ds = buzz.Dataset(
    ...     sr_work='WGS84',
    ... )

    mode 3 - Working in UTM with DXF files in WGS84 coordinates

    >>> ds = buzz.Dataset(
    ...     sr_work='EPSG:32632',
    ...     sr_fallback='WGS84',
    ... )

    mode 4 - Working in UTM with unreliable LCC input files

    >>> ds = buzz.Dataset(
    ...     sr_work='EPSG:32632',
    ...     sr_forced='EPSG:27561',
    ..  )

    .. _Scheduler:
    Scheduler
    ---------
    To handle *async rasters* living in a Dataset, a thread is to manage requests made to those
    rasters. It will start as soon as you create an *async raster* and stop when the Dataset is
    closed or collected. If one of your callbacks to be called by the scheduler raises an exception,
    the scheduler will stop and the exception will be propagated to the main thread as soon as
    possible.

    Thread-safety
    -------------
    Thread safety is one of the main concern of buzzard. Everything is thread-safe except:

    - The raster write methods
    - The vector write methods
    - The raster read methods when using the `GDAL::MEM` driver
    - The vector read methods when using the `GDAL::Memory` driver

    """

    def __init__(self, sr_work=None, sr_fallback=None, sr_forced=None,
                 analyse_transformation=True,
                 allow_none_geometry=False,
                 allow_interpolation=False,
                 max_active=np.inf,
                 debug_observers=(),
                 **kwargs):
        sr_fallback, kwargs = deprecation_pool.handle_param_renaming_with_kwargs(
            new_name='sr_fallback', old_names={'sr_implicit': '0.4.4'}, context='Dataset.__init__',
            new_name_value=sr_fallback,
            new_name_is_provided=sr_fallback is not None,
            user_kwargs=kwargs,
        )
        sr_forced, kwargs = deprecation_pool.handle_param_renaming_with_kwargs(
            new_name='sr_forced', old_names={'sr_origin': '0.4.4'}, context='Dataset.__init__',
            new_name_value=sr_forced,
            new_name_is_provided=sr_forced is not None,
            user_kwargs=kwargs,
        )
        max_active, kwargs = deprecation_pool.handle_param_renaming_with_kwargs(
            new_name='max_active', old_names={'max_activated': '0.5.0'}, context='Dataset.__init__',
            new_name_value=max_active,
            new_name_is_provided=max_active != np.inf,
            user_kwargs=kwargs,
        )
        if kwargs: # pragma: no cover
            raise TypeError("__init__() got an unexpected keyword argument '{}'".format(
                list(kwargs.keys())[0]
            ))

        mode = (sr_work is not None, sr_fallback is not None, sr_forced is not None)
        wkt_work, wkt_fallback, wkt_forced = None, None, None
        if mode == (False, False, False):
            pass
        elif mode == (True, False, False):
            success, payload = Catch(osr.GetUserInputAsWKT, nonzero_int_is_error=True)(sr_work)
            if not success:
                raise ValueError('Could not transform `sr_work` to `wkt` (gdal error: `{}`)'.format(
                    payload[1]
                ))
            wkt_work = payload
        elif mode == (True, True, False):
            success, payload = Catch(osr.GetUserInputAsWKT, nonzero_int_is_error=True)(sr_work)
            if not success:
                raise ValueError('Could not transform `sr_work` to `wkt` (gdal error: `{}`)'.format(
                    payload[1]
                ))
            wkt_work = payload
            success, payload = Catch(osr.GetUserInputAsWKT, nonzero_int_is_error=True)(sr_fallback)
            if not success:
                raise ValueError('Could not transform `sr_fallback` to `wkt` (gdal error: `{}`)'.format(
                    payload[1]
                ))
            wkt_fallback = payload
        elif mode == (True, False, True):
            success, payload = Catch(osr.GetUserInputAsWKT, nonzero_int_is_error=True)(sr_work)
            if not success:
                raise ValueError('Could not transform `sr_work` to `wkt` (gdal error: `{}`)'.format(
                    payload[1]
                ))
            wkt_work = payload
            success, payload = Catch(osr.GetUserInputAsWKT, nonzero_int_is_error=True)(sr_forced)
            if not success:
                raise ValueError('Could not transform `sr_forced` to `wkt` (gdal error: `{}`)'.format(
                    payload[1]
                ))
            wkt_forced = payload
        else:
            raise ValueError('Bad combination of `sr_*` parameters') # pragma: no cover
        del sr_work, sr_fallback, sr_forced

        if max_active < 1: # pragma: no cover
            raise ValueError('`max_active` should be greater than 1')

        allow_interpolation = bool(allow_interpolation)
        allow_none_geometry = bool(allow_none_geometry)
        analyse_transformation = bool(analyse_transformation)
        self._ds_closed = False
        self._back = BackDataset(
            wkt_work=wkt_work,
            wkt_fallback=wkt_fallback,
            wkt_forced=wkt_forced,
            analyse_transformation=analyse_transformation,
            allow_none_geometry=allow_none_geometry,
            allow_interpolation=allow_interpolation,
            max_active=max_active,
            ds_id=id(self),
            debug_observers=debug_observers,
        )
        super(Dataset, self).__init__()

    # Raster entry points *********************************************************************** **
    def open_raster(self, key, path, driver='GTiff', options=(), mode='r'):
        """Open a raster file within this Dataset under `key`. Only metadata are kept in memory.

        >>> help(GDALFileRaster)

        Parameters
        ----------
        key: hashable (like a string)
            File identifier within Dataset

            To avoid using a `key`, you may use :py:meth:`aopen_raster`
        path: string
            ..
        driver: string
            gdal driver to use when opening the file
            http://www.gdal.org/formats_list.html
        options: sequence of str
            options for gdal
        mode: one of {'r', 'w'}
            ..

        Returns
        -------
        source: GDALFileRaster
            ..

        Example
        -------
        >>> ds.open_raster('ortho', '/path/to/ortho.tif')
        >>> file_proj4 = ds.ortho.proj4_stored

        >>> ds.open_raster('dem', '/path/to/dem.tif', mode='w')
        >>> nodata_value = ds.dem.nodata

        See Also
        --------
        - :py:meth:`Dataset.aopen_raster`: To skip the `key` assigment
        - :py:func:`buzzard.open_raster`: To skip the `key` assigment and the explicit `Dataset` instanciation

        """
        # Parameter checking ***************************************************
        path = str(path)
        driver = str(driver)
        options = [str(arg) for arg in options]
        _ = conv.of_of_mode(mode)

        # Construction dispatch ************************************************
        if driver.lower() == 'mem': # pragma: no cover
            raise ValueError("Can't open a MEM raster, user create_raster")
        elif True:
            allocator = lambda: BackGDALFileRaster.open_file(
                path, driver, options, mode
            )
            prox = GDALFileRaster(self, allocator, options, mode)
        else:
            pass

        # Dataset Registering ***********************************************
        if not isinstance(key, _AnonymousSentry):
            self._register([key], prox)
        else:
            self._register([], prox)
        return prox

    def aopen_raster(self, path, driver='GTiff', options=(), mode='r'):
        """Open a raster file anonymously within this Dataset. Only metadata are kept in memory.

        See :py:meth:`~Dataset.open_raster`

        Example
        ------
        >>> ortho = ds.aopen_raster('/path/to/ortho.tif')
        >>> file_wkt = ortho.wkt_stored

        See Also
        --------
        - :py:meth:`Dataset.open_raster`: To assign a `key` to this source within the `Dataset`
        - :py:func:`buzzard.open_raster`: To skip the explicit `Dataset` instanciation

        """
        return self.open_raster(_AnonymousSentry(), path, driver, options, mode)

    def create_raster(self, key, path, fp, dtype, channel_count, channels_schema=None,
                      driver='GTiff', options=(), sr=None, ow=False, **kwargs):
        """Create a raster file and register it under `key` within this Dataset. Only metadata are
        kept in memory.

        The raster's values are initialized with `channels_schema['nodata']` or `0`.

        >>> help(GDALFileRaster)
        >>> help(GDALMemRaster)

        Parameters
        ----------
        key: hashable (like a string)
            File identifier within Dataset

            To avoid using a `key`, you may use :py:meth:`acreate_raster`
        path: string
            Anything that makes sense to GDAL:

                + A path to a file
                + An empty string when using `driver=MEM`
                + A path or an xml string when using `driver=VRT`
        fp: Footprint
            Description of the location and size of the raster to create.
        dtype: numpy type (or any alias)
            ..
        channel_count: integer
            number of channels
        channels_schema: dict or None
            Channel(s) metadata. (see `Channels schema fields` below)
        driver: string
            gdal driver to use when opening the file
            http://www.gdal.org/formats_list.html
        options: sequence of str
            options for gdal
            http://www.gdal.org/frmt_gtiff.html
        sr: string or None
            Spatial reference of the new file.

            In order not to set a spatial reference, use `None`.

            In order to set a spatial reference, use a string that can be `converted to WKT by GDAL
            <https://gdal.org/doxygen/classOGRSpatialReference.html#aec3c6a49533fe457ddc763d699ff8796>`_.
        ow: bool
            Overwrite. Whether or not to erase the existing files.

        Returns
        -------
        source: GDALFileRaster or GDALMemRaster
            The type depends on the `driver` parameter

        Example
        -------
        >>> ds.create_raster('dem_copy', 'dem_copy.tif', ds.dem.fp, ds.dsm.dtype, len(ds.dem))
        >>> array = ds.dem.get_data()
        >>> ds.dem_copy.set_data(array)

        Channel schema fields
        ---------------------
        Fields:
            'nodata': None or number
            'interpretation': None or str
            'offset': None or number
            'scale': None or number
            'mask': None or str
        Interpretation values:
            undefined, grayindex, paletteindex, redband, greenband, blueband, alphaband, hueband,
            saturationband, lightnessband, cyanband, magentaband, yellowband, blackband
        Mask values:
            all_valid, per_dataset, alpha, nodata

        Additionally:

        - A field missing or None is kept to default value.
        - A field can be passed as

            - a value: All bands are set to this value
            - a sequence of values of length `channel_count`: All bands will be set to their respective state

        Caveat
        ------
        When using the GTiff driver, specifying a `mask` or `interpretation` field may lead to unexpected results.

        See Also
        --------
        - :py:meth:`Dataset.acreate_raster`: To skip the `key` assigment
        - :py:func:`buzzard.create_raster`: To skip the `key` assigment and the explicit `Dataset` instanciation

        """

        # Deprecated parameters ************************************************
        channels_schema, kwargs = deprecation_pool.handle_param_renaming_with_kwargs(
            new_name='channels_schema', old_names={'band_schema': '0.6.0'},
            context='Dataset.create_raster',
            new_name_value=channels_schema,
            new_name_is_provided=channels_schema is not None,
            user_kwargs=kwargs,
        )
        if kwargs: # pragma: no cover
            raise TypeError("create_raster() got an unexpected keyword argument '{}'".format(
                list(kwargs.keys())[0]
            ))

        # Parameter checking ***************************************************
        ow = bool(ow)
        path = str(path)
        if not isinstance(fp, Footprint): # pragma: no cover
            raise TypeError('`fp` should be a Footprint')
        dtype = np.dtype(dtype)
        channel_count = int(channel_count)
        if channel_count <= 0:
            raise ValueError('`channel_count` should be >0')
        channels_schema = _tools.sanitize_channels_schema(channels_schema, channel_count)
        driver = str(driver)
        options = [str(arg) for arg in options]

        if sr is not None:
            success, payload = Catch(osr.GetUserInputAsWKT, nonzero_int_is_error=True)(sr)
            if not success:
                raise ValueError('Could not transform `sr` to `wkt` (gdal error: `{}`)'.format(
                    payload[1]
                ))
            wkt = payload
        else:
            wkt = None
        del sr

        if wkt is not None:
            fp = self._back.convert_footprint(fp, wkt)

        # Construction dispatch ************************************************
        if driver.lower() == 'mem':
            # TODO for 0.5.0: Check async_ is False
            prox = GDALMemRaster(
                self, fp, dtype, channel_count, channels_schema, options, wkt,
            )
        elif True:
            allocator = lambda: BackGDALFileRaster.create_file(
                path, fp, dtype, channel_count, channels_schema, driver, options, wkt, ow,
            )
            prox = GDALFileRaster(self, allocator, options, 'w')
        else:
            pass

        # Dataset Registering ***********************************************
        if not isinstance(key, _AnonymousSentry):
            self._register([key], prox)
        else:
            self._register([], prox)
        return prox

    def acreate_raster(self, path, fp, dtype, channel_count, channels_schema=None,
                       driver='GTiff', options=(), sr=None, ow=False, **kwargs):
        """Create a raster file anonymously within this Dataset. Only metadata are kept in memory.

        See :py:meth:`~Dataset.create_raster`

        Example
        -------
        >>> mask = ds.acreate_raster('mask.tif', ds.dem.fp, bool, 1, options=['SPARSE_OK=YES'])
        >>> open_options = mask.open_options

        >>> channels_schema = {
        ...     'nodata': -32767,
        ...     'interpretation': ['blackband', 'cyanband'],
        ... }
        >>> out = ds.acreate_raster('output.tif', ds.dem.fp, 'float32', 2, channels_schema)
        >>> band_interpretation = out.channels_schema['interpretation']

        See Also
        --------
        - :py:meth:`Dataset.create_raster`: To assign a `key` to this source within the `Dataset`
        - :py:func:`buzzard.create_raster`: To skip the explicit `Dataset` instanciation

        """
        return self.create_raster(_AnonymousSentry(), path, fp, dtype, channel_count, channels_schema,
                                  driver, options, sr, ow, **kwargs)

    def wrap_numpy_raster(self, key, fp, array, channels_schema=None, sr=None, mode='w', **kwargs):
        """Register a numpy array as a raster under `key` within this Dataset.

        >>> help(NumpyRaster)

        Parameters
        ----------
        key: hashable (like a string)
            File identifier within Dataset

            To avoid using a `key`, you may use :py:meth:`awrap_numpy_raster`
        fp: Footprint of shape (Y, X)
            Description of the location and size of the raster to create.
        array: ndarray of shape (Y, X) or (Y, X, C)
            ..
        channels_schema: dict or None
            Channel(s) metadata. (see `Channels schema fields` below)
        sr: string or None
            Spatial reference of the new file

            In order not to set a spatial reference, use `None`.

            In order to set a spatial reference, use a string that can be `converted to WKT by GDAL
            <https://gdal.org/doxygen/classOGRSpatialReference.html#aec3c6a49533fe457ddc763d699ff8796>`_.

        Returns
        -------
        source: NumpyRaster
            ..

        Channel schema fields
        ---------------------
        Fields:
            'nodata': None or number
            'interpretation': None or str
            'offset': None or number
            'scale': None or number
            'mask': None or str
        Interpretation values:
            undefined, grayindex, paletteindex, redband, greenband, blueband, alphaband, hueband,
            saturationband, lightnessband, cyanband, magentaband, yellowband, blackband
        Mask values:
            all_valid, per_dataset, alpha, nodata

        Additionally:

        - A field missing or None is kept to default value.
        - A field can be passed as

            - a value: All bands are set to this value
            - a sequence of values of length `channel_count`: All bands will be set to their respective state

        See Also
        --------
        - :py:meth:`Dataset.awrap_numpy_raster`: To skip the `key` assigment
        - :py:meth:`buzzard.wrap_numpy_raster`: To skip the `key` assigment and the explicit `Dataset` instanciation

        """

        # Deprecated parameters ************************************************
        channels_schema, kwargs = deprecation_pool.handle_param_renaming_with_kwargs(
            new_name='channels_schema', old_names={'band_schema': '0.6.0'},
            context='Dataset.wrap_numpy_raster',
            new_name_value=channels_schema,
            new_name_is_provided=channels_schema is not None,
            user_kwargs=kwargs,
        )
        if kwargs: # pragma: no cover
            raise TypeError("wrap_numpy_raster() got an unexpected keyword argument '{}'".format(
                list(kwargs.keys())[0]
            ))

        # Parameter checking ***************************************************
        if not isinstance(fp, Footprint): # pragma: no cover
            raise TypeError('`fp` should be a Footprint')
        array = np.asarray(array)
        if array.shape[:2] != tuple(fp.shape): # pragma: no cover
            raise ValueError('Incompatible shape between `array` and `fp`')
        if array.ndim not in [2, 3]: # pragma: no cover
            raise ValueError('Array should have 2 or 3 dimensions')
        channel_count = 1 if array.ndim == 2 else array.shape[-1]
        channels_schema = _tools.sanitize_channels_schema(channels_schema, channel_count)
        if sr is not None:
            success, payload = Catch(osr.GetUserInputAsWKT, nonzero_int_is_error=True)(sr)
            if not success:
                raise ValueError('Could not transform `sr` to `wkt` (gdal error: `{}`)'.format(
                    payload[1]
                ))
            wkt = payload
        else:
            wkt = None
        del sr
        _ = conv.of_of_mode(mode)

        if wkt is not None:
            fp = self._back.convert_footprint(fp, wkt)

        # Construction *********************************************************
        prox = NumpyRaster(self, fp, array, channels_schema, wkt, mode)

        # Dataset Registering ***********************************************
        if not isinstance(key, _AnonymousSentry):
            self._register([key], prox)
        else:
            self._register([], prox)
        return prox

    def awrap_numpy_raster(self, fp, array, channels_schema=None, sr=None, mode='w', **kwargs):
        """Register a numpy array as a raster anonymously within this Dataset.

        See Also
        --------
        - :py:meth:`Dataset.wrap_numpy_raster`: To assign a `key` to this source within the `Dataset`
        - :py:meth:`buzzard.wrap_numpy_raster`: To skip the `key` assigment and the explicit `Dataset` instanciation

        """
        return self.wrap_numpy_raster(
            _AnonymousSentry(), fp, array, channels_schema, sr, mode, **kwargs
        )

    def create_raster_recipe(
            self, key,

            # raster attributes
            fp, dtype, channel_count, channels_schema=None, sr=None,

            # callbacks running on pool
            compute_array=None, merge_arrays=buzzard.utils.concat_arrays,

            # primitives
            queue_data_per_primitive=MappingProxyType({}), convert_footprint_per_primitive=None,

            # pools
            computation_pool='cpu', merge_pool='cpu', resample_pool='cpu',

            # misc
            computation_tiles=None, max_computation_size=None,
            max_resampling_size=None, automatic_remapping=True,
            debug_observers=(),
    ):
        """

        .. warning::
            This method is not yet implemented. It exists for documentation purposes.

        Create a *raster recipe* and register it under `key` within this Dataset.

        A *raster recipe* implements the same interfaces as all other rasters, but internally it
        computes data on the fly by calling a callback. The main goal of the *raster recipes* is to
        provide a boilerplate-free interface that automatize those cumbersome tasks:

        - tiling,
        - parallelism
        - caching
        - file reads
        - resampling
        - lazy evaluation
        - backpressure prevention and
        - optimised task scheduling.

        If you are familiar with `create_cached_raster_recipe` two parameters are new here:
        `automatic_remapping` and `max_computation_size`.

        Parameters
        ----------
        key:
            see :py:meth:`Dataset.create_raster`
        fp:
            see :py:meth:`Dataset.create_raster`
        dtype:
            see :py:meth:`Dataset.create_raster`
        channel_count:
            see :py:meth:`Dataset.create_raster`
        channels_schema:
            see :py:meth:`Dataset.create_raster`
        sr:
            see :py:meth:`Dataset.create_raster`
        compute_array: callable
            see :ref:`Computation Function` below
        merge_arrays: callable
            see :ref:`Merge Function` below
        queue_data_per_primitive: dict of hashable (like a string) to a `queue_data` method pointer
            see :ref:`Primitives` below
        convert_footprint_per_primitive: None or dict of hashable (like a string) to a callable
            see :ref:`Primitives` below
        computation_pool:
            see :ref:`Pools` below
        merge_pool:
            see :ref:`Pools` below
        resample_pool:
            see :ref:`Pools` below
        computation_tiles: None or (int, int) or numpy.ndarray of Footprint
            see :ref:`Computation Tiling` below
        max_computation_size:  None or int or (int, int)
            see :ref:`Computation Tiling` below
        max_resampling_size: None or int or (int, int)
            Optionally define a maximum resampling size. If a larger resampling has to be performed,
            it will be performed tile by tile in parallel.
        automatic_remapping: bool
            see :ref:`Automatic Remapping` below
        debug_observers: sequence of object
            Entry points that observe what is happening with this raster in the Dataset's scheduler.

        Returns
        -------
        source: NocacheRasterRecipe
            ..

        .. _Computation Function:
        Computation Function
        --------------------
        The function that will map a Footprint to a numpy.ndarray. If `queue_data_per_primitive`
        is not empty, it will map a Footprint and primitive arrays to a numpy.ndarray.

        It will be called in parallel according to the `computation_pool` parameter provided at
        construction.

        The function will be called with the following positional parameters:

        - fp: Footprint of shape (Y, X)
            The location at which the pixels should be computed
        - primitive_fps: dict of hashable to Footprint
            For each primitive defined through the `queue_data_per_primitive` parameter, the input
            Footprint.
        - primitive_arrays: dict of hashable to numpy.ndarray
            For each primitive defined through the `queue_data_per_primitive` parameter, the input
            numpy.ndarray that was automatically computed.
        - raster: CachedRasterRecipe or None
            The Raster object of the ongoing computation.

        It should return either:

        - a single ndarray of shape (Y, X) if only one channel was computed
            ..
        - a single ndarray of shape (Y, X, C) if one or more channels were computed
            ..

        If `computation_pool` points to a process pool, the `compute_array` function must be
        picklable and the `raster` parameter will be None.

        .. _Computation Tiling:
        Computation Tiling
        ------------------
        You may sometimes want to have control on the Footprints that are requested to the
        `compute_array` function, for example:

        - If pixels computed by `compute_array` are long to compute, you want to tile to increase
          parallelism.
        - If the `compute_array` function scales badly in term of memory or time, you want to tile
          to reduce complexity.
        - If `compute_array` can work only on certain Footprints, you want a hard constraint on the
          set of Footprint that can be queried from `compute_array`. (This may happen with
          *convolutional neural networks*)

        To do so use the `computation_tiles` or `max_computation_size` parameter (not both).

        If `max_computation_size` is provided, a Footprint to be computed will be tiled given this
        parameter.

        If `computation_tiles` is a numpy.ndarray of Footprint, it should be a tiling of the `fp`
        parameter. Only the Footprints contained in this tiling will be asked to the
        `computation_tiles`.
        If `computation_tiles` is (int, int), a tiling will be constructed using Footprint.tile
        using those two ints.

        .. _Merge Function:
        Merge Function
        --------------
        The function that will map several pairs of Footprint/numpy.ndarray to a single
        numpy.ndarray. If the `computation_tiles` is None, it will never be called.

        It will be called in parallel according to the `merge_pool` parameter provided at
        construction.

        The function will be called with the following positional parameters:

        - fp: Footprint of shape (Y, X)
            The location at which the pixels should be computed.
        - array_per_fp: dict of Footprint to numpy.ndarray
            The pairs of Footprint/numpy.ndarray of each arrays that were computed by
            `compute_array` and that overlap with `fp`.
        - raster: CachedRasterRecipe or None
            The Raster object of the ongoing computation.

        It should return either:

        - a single ndarray of shape (Y, X) if only one channel was computed
            ..
        - a single ndarray of shape (Y, X, C) if one or more channels were computed
            ..

        If `merge_pool` points to a process pool, the `merge_array` function must be picklable and
        the `raster` parameter will be None.

        .. _Automatic Remapping:
        Automatic Remapping
        -------------------
        When creating a recipe you give a *Footprint* through the `fp` parameter. When calling your
        `compute_array` function the scheduler will only ask for slices of `fp`. This means that the
        scheduler takes care of those boilerplate steps:

        - If you request a *Footprint* on a different grid in a `get_data()` call, the scheduler
          **takes care of resampling** the outputs of your `compute*array` function.
        - If you request a *Footprint* partially or fully outside of the raster's extent, the
          scheduler will call your `compute_array` function to get the interior pixels and then
          **pad the output with nodata**.

        This system is flexible and can be deactivated by passing `automatic_remapping=False` to
        the constructor of a *NocacheRasterRecipe*, in this case the scheduler will call your
        `compute_array` function for any kind of *Footprint*; thus your function must be able to
        comply with any request.

        .. _Primitives:
        Primitives
        ----------
        The `queue_data_per_primitive` and `convert_footprint_per_primitive` parameters can be used
        to create dependencies between `dependee` *async rasters* and the *raster recipe* being
        created. The dependee/dependent relation is called primitive/derived throughout buzzard.
        A derived recipe can itself be the primitive of another raster. Pipelines of any depth and
        width can be instanciated that way.

        In `queue_data_per_primitive` you declare a `dependee` by giving it a key of your choice and
        the pointer to the `queue_data` method of `dependee` raster. You can parameterize the
        connection by *currying* the `channels`, `dst_nodata`, `interpolation` and `max_queue_size`
        parameters using `functools.partial`.

        The `convert_footprint_per_primitive` dict should contain the same keys as
        `queue_data_per_primitive`. A value in the dict should be a function that maps a Footprint
        to another Footprint. It can be used for example to request larger rectangles of primitives
        data to compute a derived array.

        e.g. If the primitive raster is an `rgb` image, and the derived raster only needs the green
        channel but with a context of 10 additional pixels on all 4 sides:

        >>> derived = ds.create_raster_recipe(
        ...     # <other parameters>
        ...     queue_data_per_primitive={'green': functools.partial(primitive.queue_data, channels=1)},
        ...     convert_footprint_per_primitive={'green': lambda fp: fp.dilate(10)},
        ... )

        .. _Pools:
        Pools
        -----
        The `*_pool` parameters can be used to select where certain computations occur. Those
        parameters can be of the following types:

        - A *multiprocessing.pool.ThreadPool*, should be the default choice.
        - A *multiprocessing.pool.Pool*, a process pool. Useful for computations that requires the
          GIL or that leaks memory.
        - `None`, to request the scheduler thread to perform the tasks itself. Should be used when
          the computation is very light.
        - A *hashable* (like a *string*), that will map to a pool registered in the *Dataset*. If
          that key is missing from the *Dataset*, a *ThreadPool* with
          `multiprocessing.cpu_count()` workers will be automatically instanciated. When the
          Dataset is closed, the pools instanciated that way will be joined.

        See Also
        --------
        - :py:meth:`Dataset.acreate_raster_recipe`: To skip the `key` assigment
        - :py:meth:`Dataset.create_raster_recipe`: For results `caching`
        - :py:meth:`Dataset.acreate_cached_raster_recipe`: To skip the `key` assigment

        """
        raise NotImplementedError()

    def create_cached_raster_recipe(
            self, key,

            # raster attributes
            fp, dtype, channel_count, channels_schema=None, sr=None,

            # callbacks running on pool
            compute_array=None, merge_arrays=buzzard.utils.concat_arrays,

            # filesystem
            cache_dir=None, ow=False,

            # primitives
            queue_data_per_primitive=MappingProxyType({}), convert_footprint_per_primitive=None,

            # pools
            computation_pool='cpu', merge_pool='cpu', io_pool='io', resample_pool='cpu',

            # misc
            cache_tiles=(512, 512), computation_tiles=None, max_resampling_size=None,
            debug_observers=()
    ):
        """Create a *cached raster recipe* and register it under `key` within this Dataset.

        Compared to a `NocacheRasterRecipe`, in a `CachedRasterRecipe` the pixels are never computed
        twice. Cache files are used to store and reuse pixels from computations. The cache can even
        be reused between python sessions.

        If you are familiar with `create_raster_recipe` four parameters are new here: `io_pool`,
        `cache_tiles`, `cache_dir` and `ow`. They are all related to file system operations.

        See `create_raster_recipe` method, since it shares most of the features:

        >>> help(CachedRasterRecipe)

        Parameters
        ----------
        key:
            see :py:meth:`Dataset.create_raster` method
        fp:
            see :py:meth:`Dataset.create_raster` method
        dtype:
            see :py:meth:`Dataset.create_raster` method
        channel_count:
            see :py:meth:`Dataset.create_raster` method
        channels_schema:
            see :py:meth:`Dataset.create_raster` method
        sr:
            see :py:meth:`Dataset.create_raster` method
        compute_array:
            see :py:meth:`Dataset.create_raster_recipe` method
        merge_arrays:
            see :py:meth:`Dataset.create_raster_recipe` method
        cache_dir: str or pathlib.Path
            Path to the directory that holds the cache files associated with this raster. If cache
            files are present, they will be reused (or erased if corrupted). If a cache file is
            needed and missing, it will be computed.
        ow: bool
            Overwrite. Whether or not to erase the old cache files contained in `cache_dir`.

            .. warning::
                not only the tiles needed (hence computed) but all buzzard cache files in
                `cache_dir` will be deleted.

        queue_data_per_primitive:
            see :py:meth:`Dataset.create_raster_recipe` method
        convert_footprint_per_primitive:
            see :py:meth:`Dataset.create_raster_recipe` method
        computation_pool:
            see :py:meth:`Dataset.create_raster_recipe` method
        merge_pool:
            see :py:meth:`Dataset.create_raster_recipe` method
        io_pool:
            see :py:meth:`Dataset.create_raster_recipe` method
        resample_pool:
            see :py:meth:`Dataset.create_raster_recipe` method
        cache_tiles: (int, int) or numpy.ndarray of Footprint
            A tiling of the `fp` parameter. Each tile will correspond to one cache file.
            if (int, int): Construct the tiling by calling Footprint.tile with this parameter
        computation_tiles:
            if None: Use the same tiling as `cache_tiles`
            else: see `create_raster_recipe` method
        max_resampling_size: None or int or (int, int)
            see :py:meth:`Dataset.create_raster_recipe` method
        debug_observers: sequence of object
            see :py:meth:`Dataset.create_raster_recipe` method

        Returns
        -------
        source: CachedRasterRecipe
            ..

        See Also
        --------
        - :py:meth:`Dataset.create_raster_recipe`: To skip the `caching`
        - :py:meth:`Dataset.acreate_cached_raster_recipe`: To skip the `key` assigment

        """
        # Parameter checking ***************************************************
        # Classic RasterSource parameters *******************
        if not isinstance(fp, Footprint): # pragma: no cover
            raise TypeError('`fp` should be a Footprint')
        dtype = np.dtype(dtype)
        channel_count = int(channel_count)
        if channel_count <= 0:
            raise ValueError('`channel_count` should be >0')
        channels_schema = _tools.sanitize_channels_schema(channels_schema, channel_count)
        if sr is not None:
            success, payload = Catch(osr.GetUserInputAsWKT, nonzero_int_is_error=True)(sr)
            if not success:
                raise ValueError('Could not transform `sr` to `wkt` (gdal error: `{}`)'.format(
                    payload[1]
                ))
            wkt = payload
        else:
            wkt = None
        del sr
        if wkt is not None:
            fp = self._back.convert_footprint(fp, wkt)

        # Callables ****************************************
        if compute_array is None:
            raise ValueError('Missing `compute_array` parameter')
        if not callable(compute_array):
            raise TypeError('`compute_array` should be callable')
        if not callable(merge_arrays):
            raise TypeError('`merge_arrays` should be callable')

        # Primitives ***************************************
        if convert_footprint_per_primitive is None:
            convert_footprint_per_primitive = {
                name: (lambda fp: fp)
                for name in queue_data_per_primitive.keys()
            }

        if queue_data_per_primitive.keys() != convert_footprint_per_primitive.keys():
            err = 'There should be the same keys in `queue_data_per_primitive` and '
            err += '`convert_footprint_per_primitive`.'
            if queue_data_per_primitive.keys() - convert_footprint_per_primitive.keys():
                err += '\n{} are missing from `convert_footprint_per_primitive`.'.format(
                    queue_data_per_primitive.keys() - convert_footprint_per_primitive.keys()
                )
            if convert_footprint_per_primitive.keys() - queue_data_per_primitive.keys():
                err += '\n{} are missing from `queue_data_per_primitive`.'.format(
                    convert_footprint_per_primitive.keys() - queue_data_per_primitive.keys()
                )
            raise ValueError(err)

        primitives_back = {}
        primitives_kwargs = {}
        for name, met in queue_data_per_primitive.items():
            primitives_back[name], primitives_kwargs[name] = _tools.shatter_queue_data_method(met, name)
            if primitives_back[name].back_ds is not self._back:
                raise ValueError('The `{}` primitive comes from another Dataset'.format(
                    name
                ))

        for name, func in convert_footprint_per_primitive.items():
            if not callable(func):
                raise TypeError('convert_footprint_per_primitive[{}] should be callable'.format(
                    name
                ))

        # Pools ********************************************
        computation_pool = self._back.pools_container._normalize_pool_parameter(
            computation_pool, 'computation_pool'
        )
        merge_pool = self._back.pools_container._normalize_pool_parameter(
            merge_pool, 'merge_pool'
        )
        io_pool = self._back.pools_container._normalize_pool_parameter(
            io_pool, 'io_pool'
        )
        resample_pool = self._back.pools_container._normalize_pool_parameter(
            resample_pool, 'resample_pool'
        )

        # Tilings ******************************************
        if isinstance(cache_tiles, np.ndarray) and cache_tiles.dtype == np.object:
            if not _tools.is_tiling_covering_fp(
                    cache_tiles, fp,
                    allow_outer_pixels=False, allow_overlapping_pixels=False,
            ):
                raise ValueError("`cache_tiles` should be a tiling of raster's Footprint, " +\
                                "without overlap, with `boundary_effect='shrink'`"
                )
        else:
            # Defer the parameter checking to fp.tile
            cache_tiles = fp.tile(cache_tiles, 0, 0, boundary_effect='shrink')

        if computation_tiles is None:
            computation_tiles = cache_tiles
        elif isinstance(computation_tiles, np.ndarray) and computation_tiles.dtype == np.object:
            if not _tools.is_tiling_covering_fp(
                    cache_tiles, fp,
                    allow_outer_pixels=True, allow_overlapping_pixels=True,
            ):
                raise ValueError("`computation_tiles` should be a tiling covering raster's Footprint")
        else:
            # Defer the parameter checking to fp.tile
            computation_tiles = fp.tile(computation_tiles, 0, 0, boundary_effect='shrink')

        # Misc *********************************************
        if max_resampling_size is not None:
            max_resampling_size = int(max_resampling_size)
            if max_resampling_size <= 0:
                raise ValueError('`max_resampling_size` should be >0')

        if cache_dir is None:
            raise ValueError('Missing `cache_dir` parameter')
        if not isinstance(cache_dir, (str, pathlib.Path)):
            raise TypeError('cache_dir should be a string')
        cache_dir = str(cache_dir)
        overwrite = bool(ow)
        del ow

        # Construction *********************************************************
        prox = CachedRasterRecipe(
            self,
            fp, dtype, channel_count, channels_schema, wkt,
            compute_array, merge_arrays,
            cache_dir, overwrite,
            primitives_back, primitives_kwargs, convert_footprint_per_primitive,
            computation_pool, merge_pool, io_pool, resample_pool,
            cache_tiles, computation_tiles,
            max_resampling_size,
            debug_observers,
        )

        # Dataset Registering ***********************************************
        if not isinstance(key, _AnonymousSentry):
            self._register([key], prox)
        else:
            self._register([], prox)
        return prox

    def acreate_cached_raster_recipe(
            self,

            # raster attributes
            fp, dtype, channel_count, channels_schema=None, sr=None,

            # callbacks running on pool
            compute_array=None, merge_arrays=buzzard.utils.concat_arrays,

            # filesystem
            cache_dir=None, ow=False,

            # primitives
            queue_data_per_primitive=MappingProxyType({}), convert_footprint_per_primitive=None,

            # pools
            computation_pool='cpu', merge_pool='cpu', io_pool='io', resample_pool='cpu',

            # misc
            cache_tiles=(512, 512), computation_tiles=None, max_resampling_size=None,
            debug_observers=()
    ):
        """Create a cached raster reciped anonymously within this Dataset.

        See Dataset.create_cached_raster_recipe

        See Also
        --------
        - :py:meth:`Dataset.create_raster_recipe`: To skip the `caching`
        - :py:meth:`Dataset.create_cached_raster_recipe`: To assign a `key` to this source within the `Dataset`

        """
        return self.create_cached_raster_recipe(
            _AnonymousSentry(),
            fp, dtype, channel_count, channels_schema, sr,
            compute_array, merge_arrays,
            cache_dir, ow,
            queue_data_per_primitive, convert_footprint_per_primitive,
            computation_pool, merge_pool, io_pool, resample_pool,
            cache_tiles, computation_tiles, max_resampling_size,
            debug_observers,
        )

    # Vector entry points *********************************************************************** **
    def open_vector(self, key, path, layer=None, driver='ESRI Shapefile', options=(), mode='r'):
        """Open a vector file within this Dataset under `key`. Only metadata are kept in memory.

        >>> help(GDALFileVector)

        Parameters
        ----------
        key: hashable (like a string)
            File identifier within Dataset

            To avoid using a `key`, you may use :py:meth:`aopen_vector`
        path: string
            ..
        layer: None or int or string
            ..
        driver: string
            ogr driver to use when opening the file
            http://www.gdal.org/ogr_formats.html
        options: sequence of str
            options for ogr
        mode: one of {'r', 'w'}
            ..

        Returns
        -------
        source: GDALFileVector
            ..

        Example
        -------
        >>> ds.open_vector('trees', '/path/to.shp')
        >>> feature_count = len(ds.trees)

        >>> ds.open_vector('roofs', '/path/to.json', driver='GeoJSON', mode='w')
        >>> fields_list = ds.roofs.fields

        See Also
        --------
        - :py:meth:`Dataset.aopen_vector`: To skip the `key` assigment
        - :py:func:`buzzard.open_vector`: To skip the `key` assigment and the explicit `Dataset` instanciation

        """
        # Parameter checking ***************************************************
        path = str(path)
        if layer is None:
            layer = 0
        elif np.all(np.isreal(layer)):
            layer = int(layer)
        else:
            layer = str(layer)
        driver = str(driver)
        options = [str(arg) for arg in options]
        _ = conv.of_of_mode(mode)

        # Construction dispatch ************************************************
        if driver.lower() == 'memory': # pragma: no cover
            raise ValueError("Can't open a MEMORY vector, user create_vector")
        elif True:
            allocator = lambda: BackGDALFileVector.open_file(
                path, layer, driver, options, mode
            )
            prox = GDALFileVector(self, allocator, options, mode)
        else:
            pass

        # Dataset Registering ***********************************************
        if not isinstance(key, _AnonymousSentry):
            self._register([key], prox)
        else:
            self._register([], prox)
        return prox

    def aopen_vector(self, path, layer=None, driver='ESRI Shapefile', options=(), mode='r'):
        """Open a vector file anonymously within this Dataset. Only metadata are kept in memory.

        See :py:meth:`~Dataset.open_vector`

        Example
        -------
        >>> trees = ds.aopen_vector('/path/to.shp')
        >>> features_bounds = trees.bounds

        See Also
        --------
        - :py:meth:`Dataset.open_vector`: To assign a `key` to this source within the `Dataset`
        - :py:func:`buzzard.open_vector`: To skip the `key` assigment and the explicit `Dataset` instanciation

        """
        return self.open_vector(_AnonymousSentry(), path, layer, driver, options, mode)

    def create_vector(self, key, path, type, fields=(), layer=None,
                      driver='ESRI Shapefile', options=(), sr=None, ow=False):
        """Create an empty vector file and register it under `key` within this Dataset. Only metadata
        are kept in memory.

        >>> help(GDALFileVector)
        >>> help(GDALMemoryVector)

        Parameters
        ----------
        key: hashable (like a string)
            File identifier within Dataset

            To avoid using a `key`, you may use :py:meth:`acreate_vector`
        path: string
            Anything that makes sense to GDAL:

                + A path to a file
                + An empty string when using `driver=Memory`
        type: string
            name of a wkb geometry type, without the `wkb` prefix.

            list: http://www.gdal.org/ogr__core_8h.html#a800236a0d460ef66e687b7b65610f12a

        fields: sequence of dict
            Attributes of fields, one dict per field. (see :ref:`Field Attributes` below)
        layer: None or string
            ..
        driver: string
            ogr driver to use when opening the file
            http://www.gdal.org/ogr_formats.html
        options: sequence of str
            options for ogr
        sr: string or None
            Spatial reference of the new file

            In order not to set a spatial reference, use `None`.

            In order to set a spatial reference, use a string that can be `converted to WKT by GDAL
            <https://gdal.org/doxygen/classOGRSpatialReference.html#aec3c6a49533fe457ddc763d699ff8796>`_.
        ow: bool
            Overwrite. Whether or not to erase the existing files.

        Returns
        -------
        source: GDALFileVector or GDALMemoryVector
            The type depends on the `driver` parameter

        Example
        -------
        >>> ds.create_vector('lines', '/path/to.shp', 'linestring')
        >>> geometry_type = ds.lines.type
        >>> ds.lines.insert_data([[0, 0], [1, 1], [1, 2]])

        >>> fields = [
            {'name': 'name', 'type': str},
            {'name': 'count', 'type': 'int32'},
            {'name': 'area', 'type': np.float64, 'width': 5, precision: 18},
            {'name': 'when', 'type': np.datetime64},
        ]
        >>> ds.create_vector('zones', '/path/to.shp', 'polygon', fields)
        >>> field0_type = ds.zones.fields[0]['type']
        >>> ds.zones.insert_data(shapely.geometry.box(10, 10, 15, 15))

        .. _Field Attributes:
        Field Attributes
        ----------------
        Attributes:

        - "name": string
        - "type": string (see :ref:`Field Types` below)
        - "precision": int
        - "width": int
        - "nullable": bool
        - "default": same as `type`

        An attribute missing or None is kept to default value.

        .. _Field Types:
        Field Types
        -----------
        +---------------+------------------------------------------------------------------------+
        | Type          | Type names                                                             |
        +===============+========================================================================+
        | Binary        | "binary", bytes, np.bytes\_, aliases of np.bytes\_                     |
        +---------------+------------------------------------------------------------------------+
        | Date          | "date"                                                                 |
        +---------------+------------------------------------------------------------------------+
        | DateTime      | "datetime", datetime.datetime, np.datetime64, aliases of np.datetime64 |
        +---------------+------------------------------------------------------------------------+
        | Time          | "time"                                                                 |
        +---------------+------------------------------------------------------------------------+
        | Integer       | "integer" np.int32, aliases of np.int32                                |
        +---------------+------------------------------------------------------------------------+
        | Integer64     | "integer64", int, np.int64, aliases of np.int64                        |
        +---------------+------------------------------------------------------------------------+
        | Real          | "real", float, np.float64, aliases of np.float64                       |
        +---------------+------------------------------------------------------------------------+
        | String        | "string", str, np.str\_, aliases of np.str\_                           |
        +---------------+------------------------------------------------------------------------+
        | Integer64List | "integer64list", "int list"                                            |
        +---------------+------------------------------------------------------------------------+
        | IntegerList   | "integerlist"                                                          |
        +---------------+------------------------------------------------------------------------+
        | RealList      | "reallist", "float list"                                               |
        +---------------+------------------------------------------------------------------------+

        See Also
        --------
        - :py:meth:`Dataset.acreate_vector`: To skip the `key` assigment
        - :py:func:`buzzard.create_vector`: To skip the `key` assigment and the explicit `Dataset` instanciation

        """
        type_ = type
        del type

        # Parameter checking ***************************************************
        path = str(path)
        type_ = conv.str_of_wkbgeom(conv.wkbgeom_of_str(type_))
        fields = _tools.normalize_fields_defn(fields)
        if layer is None:
            layer = '.'.join(os.path.basename(path).split('.')[:-1])
        else:
            layer = str(layer)
        driver = str(driver)
        options = [str(arg) for arg in options]
        ow = bool(ow)
        if sr is None:
            wkt = None
        else:
            success, payload = Catch(osr.GetUserInputAsWKT, nonzero_int_is_error=True)(sr)
            if not success:
                raise ValueError('Could not transform `sr` to `wkt` (gdal error: `{}`)'.format(
                    payload[1]
                ))
            wkt = payload

        # Construction dispatch ************************************************
        if driver.lower() == 'memory':
            allocator = lambda: BackGDALFileVector.create_file(
                '', type_, fields, layer, 'Memory', options, wkt, False,
            )
            prox = GDALMemoryVector(self, allocator, options)
        elif True:
            allocator = lambda: BackGDALFileVector.create_file(
                path, type_, fields, layer, driver, options, wkt, ow
            )
            prox = GDALFileVector(self, allocator, options, 'w')
        else:
            pass

        # Dataset Registering ***********************************************
        if not isinstance(key, _AnonymousSentry):
            self._register([key], prox)
        else:
            self._register([], prox)
        return prox

    def acreate_vector(self, path, type, fields=(), layer=None,
                       driver='ESRI Shapefile', options=(), sr=None, ow=False):
        """Create a vector file anonymously within this Dataset. Only metadata are kept in memory.

        See :py:meth:`~Dataset.create_vector`

        Example
        -------
        >>> lines = ds.acreate_vector('/path/to.shp', 'linestring')
        >>> file_proj4 = lines.proj4_stored

        See Also
        --------
        - :py:meth:`Dataset.create_vector`: To assign a `key` to this source within the `Dataset`
        - :py:func:`buzzard.create_vector`: To skip the `key` assigment and the explicit `Dataset` instanciation

        """
        return self.create_vector(_AnonymousSentry(), path, type, fields, layer,
                                  driver, options, sr, ow)

    # Cleanup *********************************************************************************** **
    def __del__(self):
        if not self._ds_closed:
            self.close()

    @property
    def close(self):
        """Close the Dataset with a call or a context management.
        The `close` attribute returns an object that can be both called and used in a with statement

        The Dataset can be closed manually or automatically when garbage collected, it is safer
        to do it manually.

        The internal steps are:

        - Stopping the scheduler
        - Joining the mp.Pool that have been automatically allocated
        - Closing all sources

        Examples
        --------
        >>> ds = buzz.Dataset()
        ... # code...
        ... ds.close()

        >>> with buzz.Dataset().close as ds
        ...     # code...

        Caveat
        ------
        When using a scheduler, some memory leaks may still occur after closing a Dataset.
        Possible origins:

        - https://bugs.python.org/issue34172 (update your python to >=3.6.7)
        - Gdal cache not flushed (not a leak)
        - The gdal version
        - https://stackoverflow.com/a/1316799 (not a leak)
        - Some unknown leak in the python `threading` or `multiprocessing` standard library
        - Some unknown library leaking memory on the `C` side
        - Some unknown library storing data in global variables

        You can use a `debug_observer` with an `on_object_allocated` method to track large objects
        allocated in the scheduler. It will likely not be the source of the problem. If you
        even find a source of leaks please contact the buzzard team.
        https://github.com/earthcube-lab/buzzard/issues

        """
        if self._ds_closed:
            raise RuntimeError("Dataset already closed")

        def _close():
            if self._ds_closed:
                raise RuntimeError("Dataset already closed")
            self._ds_closed = True

            # Tell scheduler to stop, wait until it is done
            self._back.stop_scheduler()

            # Safely release all resources
            self._back.pools_container._close()
            for source in list(self._keys_of_source.keys()):
                source.close()

        return _CloseRoutine(self, _close)

    # Source infos ******************************************************************************* **
    def __getitem__(self, key):
        """Retrieve a source from its key"""
        return self._source_of_key[key]

    def __contains__(self, item):
        """Is key or source registered in Dataset"""
        if isinstance(item, ASource):
            return item in self._keys_of_source
        return item in self._source_of_key

    def items(self):
        """Generate the pair of (keys_of_source, source) for all proxies"""
        for source, keys in self._keys_of_source.items():
            yield list(keys), source

    def keys(self):
        """Generate all source keys"""
        for source, keys in self._keys_of_source.items():
            for key in keys:
                yield key

    def values(self):
        """Generate all proxies"""
        for source, _ in self._keys_of_source.items():
            yield source

    def __len__(self):
        """Retrieve source count registered within this Dataset"""
        return len(self._keys_of_source)

    # Spatial reference getters ***************************************************************** **
    @property
    def proj4(self):
        """Dataset's work spatial reference in WKT proj4.
        Returns None if `mode 1`.
        """
        if self._back.wkt_work is None:
            return None
        return osr.SpatialReference(self._back.wkt_work).ExportToProj4()

    @property
    def wkt(self):
        """Dataset's work spatial reference in WKT format.
        Returns None if `mode 1`.
        """
        return self._back.wkt_work

    # Activation mechanisms ********************************************************************* **
    @property
    def active_count(self):
        """Count how many driver objects are currently active"""
        return self._back.active_count()

    def activate_all(self):
        """Activate all deactivable proxies.
        May raise an exception if the number of sources is greater than `max_activated`
        """
        proxs = [
            prox
            for prox in self._keys_of_source.keys()
            if isinstance(prox, APooledEmissary)
        ]
        total = len(proxs)

        if self._back.max_active < total:
            raise RuntimeError("Can't activate all pooled sources at the same time: {} pooled sources and max_activated is {}".format(
                total, self._back.max_active,
            ))

        # Hacky implementation to get the expected behavior
        # TODO: Implement that routine in the back driver pool. Is it possible? We need to call `.activate`
        i = 0
        for prox in itertools.cycle(proxs):
            if i == total:
                break
            if not prox.active:
                prox.activate()
                i = 1
            else:
                i += 1

    def deactivate_all(self):
        """Deactivate all deactivable proxies. Useful to flush all files to disk"""
        for prox in self._keys_of_source.keys():
            if prox.active:
                prox.deactivate()

    # Pools infos ******************************************************************************* **
    @property
    def pools(self):
        """Get the Pool Container.

        >>> help(PoolsContainer)

        """
        return self._back.pools_container

    # Deprecation ******************************************************************************* **
    open_araster = deprecation_pool.wrap_method(
        aopen_raster,
        '0.4.4'
    )
    create_araster = deprecation_pool.wrap_method(
        acreate_raster,
        '0.4.4'
    )
    open_avector = deprecation_pool.wrap_method(
        aopen_vector,
        '0.4.4'
    )
    create_avector = deprecation_pool.wrap_method(
        acreate_vector,
        '0.4.4'
    )

    # The end *********************************************************************************** **
    # ******************************************************************************************* **

if sys.version_info < (3, 6):
    # https://www.python.org/dev/peps/pep-0487/
    for k, v in Dataset.__dict__.items():
        if hasattr(v, '__set_name__'):
            v.__set_name__(Dataset, k)

def open_raster(*args, **kwargs):
    """Shortcut for `Dataset().aopen_raster`

    >>> help(Dataset.open_raster)

    See Also
    --------
    - :py:func:`Dataset.open_raster`
    - :py:meth:`Dataset.aopen_raster`

    """
    return Dataset().aopen_raster(*args, **kwargs)

def create_raster(*args, **kwargs):
    """Shortcut for `Dataset().acreate_raster`

    >>> help(Dataset.create_raster)

    See Also
    --------
    - :py:func:`Dataset.create_raster`
    - :py:meth:`Dataset.acreate_raster`

    """
    return Dataset().acreate_raster(*args, **kwargs)

def open_vector(*args, **kwargs):
    """Shortcut for `Dataset().aopen_vector`

    >>> help(Dataset.open_vector)

    See Also
    --------
    - :py:func:`Dataset.open_vector`
    - :py:meth:`Dataset.aopen_vector`

    """
    return Dataset().aopen_vector(*args, **kwargs)

def create_vector(*args, **kwargs):
    """Shortcut for `Dataset().acreate_vector`

    >>> help(Dataset.create_vector)

    See Also
    --------
    - :py:func:`Dataset.create_vector`
    - :py:meth:`Dataset.acreate_vector`

    """
    return Dataset().acreate_vector(*args, **kwargs)

def wrap_numpy_raster(*args, **kwargs):
    """Shortcut for `Dataset().awrap_numpy_raster`

    >>> help(Dataset.wrap_numpy_raster)

    See Also
    --------
    - :py:func:`Dataset.wrap_numpy_raster`
    - :py:meth:`Dataset.awrap_numpy_raster`

    """
    return Dataset().awrap_numpy_raster(*args, **kwargs)

_CloseRoutine = type('_CloseRoutine', (_tools.CallOrContext,), {
    '__doc__': Dataset.close.__doc__,

})

DataSource = deprecation_pool.wrap_class(Dataset, 'DataSource', '0.6.0')

class _AnonymousSentry(object):
    """Sentry object used to instanciate anonymous proxies"""



from buzzard._a_emissary_raster import AEmissaryRaster, ABackEmissaryRaster
from buzzard._a_pooled_emissary import APooledEmissary, ABackPooledEmissary

class APooledEmissaryRaster(APooledEmissary, AEmissaryRaster):
    """Base abstract class defining the common behavior of all rasters that can deactivate and
    reactivate their underlying driver at will.

    Features Defined
    ----------------
    None
    """
    pass

class ABackPooledEmissaryRaster(ABackPooledEmissary, ABackEmissaryRaster):
    """Implementation of APooledEmissaryRaster's specifications"""
    pass



import sys
import uuid

import numpy as np

class Msg(object):
    """Message exchanged:
    - between two actors (receiver method prefixed by `receive_`)
    - from scheduler to actor (receiver method prefixed by `ext_receive_`)
    - from user's thread to scheduler to actor (receiver method also prefixed by `ext_receive_`)
    """
    __slots__ = ['address', 'title', 'args']

    def __init__(self, address, title, *args):
        """
        Parameters
        ----------
        address: str
            Destination of the message.
            Relative or Absolute address
        title: str
            Method to call
        *args: sequence of object
            Arguments for the method call
        """
        self.address = address
        self.title = title
        self.args = args

    def __str__(self): # pragma: no cover
        if self.address[0] == '/':
            b = _COLOR_PER_CLASSNAME.get(self.address.split('/')[2], '\033[37m')
        else:
            b = _COLOR_PER_CLASSNAME.get(self.address, '\033[37m')

        def _dump_param(a):

            # if isinstance(a, str):
                # return '"{}"'.format(a)
            if isinstance(a, (int, uuid.UUID)):
                return str(a)
            elif isinstance(a, (np.ndarray)):
                return '{}{}'.format(a.dtype, a.shape)
            elif isinstance(a, (tuple, list)):
                types = list(set(type(b).__name__ for b in a))
                types = '|'.join(types)
                s = '[{}]*{}'.format(types, len(a))
                t = '[{}]'.format(', '.join(map(_dump_param, a)))
                return min([t, s], key=len)
            elif type(a).__name__ == 'CachedQueryInfos':
                return 'qi:{:#x}'.format(id(a))
            elif type(a).__name__ == 'Footprint':
                return 'Footprint{:#18x}'.format(hash(a) % ((sys.maxsize + 1) * 2))
            else:
                s = type(a).__name__
                t = str(a)
                return min([t, s], key=len)

        return '{u}{b}{letter}{}{z}.{}({}){u}{b}{letter}{z}'.format(
            self.address,
            self.title,
            ', '.join(_dump_param(a) for a in self.args),
            letter=u"\u2709",
            u='\033[4m',
            b=b,
            z='\033[0m'
        )

class DroppableMsg(Msg):
    pass

class AgingMsg(Msg):
    def __init__(self, address, title, id_args, other_args):
        self.id_args = id_args
        super().__init__(address, title, *(list(id_args) + list(other_args)))

_COLOR_PER_CLASSNAME = {
    'TopLevel': '\033[37m',
    'GlobalPrioritiesWatcher': '\033[37m',

    'WaitingRoom': '\033[36m',
    'WorkingRoom': '\033[36m',

    'ProductionGate': '\033[32m',
    'Producer': '\033[32m',
    'Resampler': '\033[32m',
    'CacheExtractor': '\033[32m',
    'Reader': '\033[32m',

    'QueriesHandler': '\033[33m',
    'CacheSupervisor': '\033[33m',
    'FileChecker': '\033[33m',

    'ComputationAccumulator': '\033[35m',
    'Computer': '\033[35m',
    'Merger': '\033[35m',
    'Writer': '\033[35m',
    'ComputationGate1': '\033[35m',
    'ComputationGate2': '\033[35m',
}



"""
Run all unit tests
"""

import os

import pytest

def _m():
    directory = os.path.join(os.path.dirname(os.path.abspath(__file__)))
    pytest.main([directory])

if __name__ == '__main__':
    _m()




from buzzard._dataset_back_conversions import BackDatasetConversionsMixin
from buzzard._dataset_back_activation_pool import BackDatasetActivationPoolMixin
from buzzard._dataset_back_scheduler import BackDatasetSchedulerMixin
from buzzard._dataset_pools_container import PoolsContainer

class BackDataset(BackDatasetConversionsMixin,
                     BackDatasetActivationPoolMixin,
                     BackDatasetSchedulerMixin):
    """Backend of the Dataset, referenced by backend proxies
    Implements activation (pooling) and conversion methods"""

    def __init__(self, allow_none_geometry, allow_interpolation, **kwargs):
        self.allow_interpolation = allow_interpolation
        self.allow_none_geometry = allow_none_geometry
        self.pools_container = PoolsContainer()
        super(BackDataset, self).__init__(**kwargs)



from buzzard._actors.message import Msg

class ActorComputationGate1(object):
    """Actor that takes care of delaying the computation of a cache file until needed soon by
    the query. It receives resquests to compute `cache footprints`, it outputs requests to compute
    `compute footprints`.
    """

    def __init__(self, raster):
        self._raster = raster
        self._queries = {}
        self._alive = True
        self.address = '/Raster{}/ComputationGate1'.format(self._raster.uid)

    @property
    def alive(self):
        return self._alive

    # ******************************************************************************************* **
    def receive_compute_those_cache_files(self, qi):
        """Receive message: The collection started for the computation of the missing cache files,
        allow computation as soon as the cache file is needed soon by query

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        """
        msgs = []

        qicc = qi.cache_computation
        assert qicc is not None

        if qi in self._queries:
            # `receive_output_queue_update` happened before this call
            q = self._queries[qi]
        else:
            q = _Query()
            self._queries[qi] = q
        msgs += self._allow(qi, q)
        return msgs

    def receive_output_queue_update(self, qi, produced_count, queue_size):
        """Receive message: The output queue of a query changed in size.

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        produced_count: int
            How many arrays were pushed in the output queue
        queue_size: int
            How many arrays are currently available in the output queue
        """
        msgs = []

        pulled_count = produced_count - queue_size
        qicc = qi.cache_computation

        if qi in self._queries:
            assert pulled_count >= self._queries[qi].pulled_count, (
                'new pulled_count:{} '.format(pulled_count) +
                'previous pulled_count:{}'.format(self._queries[qi].pulled_count)
            )
            assert produced_count >= self._queries[qi].produced_count, (
                'new produced_count:{} '.format(produced_count) +
                'previous produced_count:{}'.format(self._queries[qi].produced_count)
            )

        if produced_count == qi.produce_count:
            # Query finished
            if qi in self._queries:
                # Either qicc is None (no computation needed)
                #  or allowed count == produced_count (all computations allowed here)
                #  or allowed count <  produced_count (some other query computed the arrays)
                del self._queries[qi]
        else:
            if qi in self._queries:
                q = self._queries[qi]
            else:
                q = _Query()
                self._queries[qi] = q
            q.pulled_count = pulled_count
            if qicc is not None:
                msgs += self._allow(qi, q)

            q.produced_count = produced_count

        return msgs

    def receive_cancel_this_query(self, qi):
        """Receive message: One query was dropped

        Parameters
        ----------
        qi: _actors.cached.query_infos.QueryInfos
        """
        if qi in self._queries:
            del self._queries[qi]
        return []

    def receive_die(self):
        """Receive message: The raster was killed"""
        assert self._alive
        self._alive = False
        self._queries.clear()
        self._raster = None
        return []

    # ******************************************************************************************* **
    @staticmethod
    def _allow(qi, q):
        msgs = []
        qicc = qi.cache_computation

        max_prod_idx_allowed = q.pulled_count + qi.max_queue_size - 1
        i = q.allowed_count
        while True:
            # list_of_compute_fp being sorted by priority, `min_prod_idx` is increasing between loops

            if i == len(qicc.list_of_compute_fp):
                break
            compute_fp = qicc.list_of_compute_fp[i]
            min_prod_idx = qicc.dict_of_min_prod_idx_per_compute_fp[compute_fp]
            if min_prod_idx > max_prod_idx_allowed:
                break
            msgs += [Msg(
                'ComputationGate2', 'compute_this_array', qi, i,
            )]
            i += 1
        q.allowed_count = i

        return msgs

    # ******************************************************************************************* **

class _Query(object):
    def __init__(self):
        self.pulled_count = 0
        self.allowed_count = 0
        self.produced_count = 0




# pylint: disable=redefined-outer-name

import numpy as np
import numpy.random as npr
import pytest
import shapely.geometry as sg

import buzzard as buzz

_RANDOM_ORDERING_COUNT = 50

_GRID = """---------------------
---------a-----------
--000000000000000----
--0f--b--------g0----
--0-11111-22222-03---
--0-1---1-2-i-2-0-44-
-90-1-j-1-2-k-2-0-44-
--0-1l--1-2---2c0----
--0-11111d22222-05---
--0e-----------h0----
--000000000000000mmm-
--nnnn--88--6----m-m-
--n--n--88---7---mmm-
--n--n--88-----------
--nnnn------oooo-----
------------o--o-----
------------oooo-----
---------------------"""

def _dump_mask(mask):
    print('\n'.join(
        ' '.join(
            'X' if val else '-'
            for val in line
        )
        for line in mask
    ))

@pytest.fixture(scope='module')
def chars_grid():
    return np.asarray([list(line) for line in _GRID.split('\n')])

@pytest.fixture(scope='module')
def geometries(chars_grid):
    print(chars_grid)
    chars_set = {c for c in _GRID if c not in ['\n', '-']}

    def _foorprint_of_char(c):
        ys, xs = np.where(chars_grid == c)
        minx = xs.min()
        maxx = xs.max()
        miny = ys.min()
        maxy = ys.max()
        rsize = maxx - minx + 1, maxy - miny + 1
        fp = buzz.Footprint(tl=(minx, -miny), rsize=rsize, size=rsize)
        return fp

    def _geometry_of_footprint(fp):
        if fp.rarea != fp.rlength:
            return fp.poly - fp.erode(1).poly
        else:
            return fp.poly

    return list(map(_geometry_of_footprint, map(_foorprint_of_char, chars_set)))

@pytest.fixture(scope='module')
def truth(chars_grid):
    return chars_grid != '-'

@pytest.fixture(scope='module')
def fullfp(truth):
    rsize = np.flipud(truth.shape)
    return buzz.Footprint(tl=(0, 0), rsize=rsize, size=rsize)

def seeds():
    rng = npr.RandomState(42)
    return list(rng.randint(0, 10000, _RANDOM_ORDERING_COUNT))

@pytest.fixture(params=seeds())
def shuffled_geometries(request, geometries):
    seed = request.param
    geometries = list(geometries)
    npr.RandomState(seed).shuffle(geometries)
    return geometries

def test_burn(fullfp, shuffled_geometries, truth):
    geoms = shuffled_geometries
    res = fullfp.burn_polygons(geoms)
    if (res != truth).any():
        print('REF **************************************************')
        _dump_mask(truth)
        print('RES **************************************************')
        _dump_mask(res)
        assert False

def test_find(fullfp, geometries, truth):
    multipoly_ref = sg.MultiPolygon(geometries)
    geometries_test = fullfp.find_polygons(truth)
    multipoly_test = sg.MultiPolygon(geometries_test)
    assert (multipoly_ref ^ multipoly_test).is_empty




# pylint: disable=redefined-outer-name

from __future__ import division, print_function
import itertools

import numpy as np
import pytest
from affine import Affine

import buzzard as buzz
from buzzard.test.tools import eq, fpeq, eqall
from buzzard.test import make_tile_set

LETTERS = "ABCDEFGHI"


@pytest.fixture(scope='module')
def fps():
    """
    See make_tile_set
    A B C
    D E F
    G H I
    """
    return make_tile_set.make_tile_set(3, [0.1, -0.1])

@pytest.fixture(scope='module')
def fps1px():
    """
    See make_tile_set
    A B C
    D E F
    G H I
    """
    return make_tile_set.make_tile_set(3, [1, -1], (1, -1))

def test_size_accessors_spatial(fps):
    assert eq(fps.A.size, fps.B.size, fps.C.size, fps.D.size, fps.E.size,
              fps.F.size, fps.G.size, fps.H.size, fps.I.size,)
    assert eq(fps.AE.size, fps.BF.size, fps.DH.size, fps.EI.size)
    assert eq(fps.AH.size, fps.BI.size)
    assert eq(fps.AG.h, fps.BH.h, fps.CI.h, fps.CI.height,)
    assert eq(fps.AB.w, fps.AE.w, fps.AH.w, fps.HI.w, fps.HI.width, )


def test_size_accessors_raster(fps):
    assert eq(fps.A.rtl, [0, 0], [fps.A.rtlx, fps.A.rtly])
    assert eq(fps.A.rtl, [0, 0])
    assert eq(fps.A.rsize, fps.B.rsize, fps.C.rsize, fps.D.rsize, fps.E.rsize,
              fps.F.rsize, fps.G.rsize, fps.H.rsize, fps.I.rsize, )
    assert eq(fps.AE.rsize, fps.BF.rsize, fps.DH.rsize, fps.EI.rsize, )
    assert eq(fps.AH.rsize, fps.BI.rsize, )
    assert eq(fps.AG.rh, fps.BH.rh, fps.CI.rh, fps.CI.rheight, )
    assert eq(fps.AB.rw, fps.AE.rw, fps.AH.rw, fps.HI.rw, fps.HI.rwidth, )

    assert eq(fps.A.shape, fps.B.shape, fps.C.shape, fps.D.shape, fps.E.shape,
              fps.F.shape, fps.G.shape, fps.H.shape, fps.I.shape, )
    assert eq(fps.AE.shape, fps.BF.shape, fps.DH.shape, fps.EI.shape, )
    assert eq(fps.AH.shape, fps.BI.shape, )


def test_coordinates_accessors_spatial_corners(fps):
    buzz.Footprint(
        gt=fps.A.gt, rsize=fps.A.rsize
    )
    assert eq(fps.E.tl, fps.B.bl, fps.A.br, fps.D.tr, )
    assert eq(fps.E.bl, fps.D.br, fps.G.tr, fps.H.tl, )
    assert eq(fps.E.br, fps.H.tr, fps.I.tl, fps.F.bl, )
    assert eq(fps.E.tr, fps.F.tl, fps.C.bl, fps.B.br, )
    for letter in LETTERS:
        those_tl = [v.tl for k, v in fps.items() if k.startswith(letter)]
        assert eq(*those_tl)
        those_br = [v.br for k, v in fps.items() if k.endswith(letter)]
        assert eq(*those_br)


def test_coordinates_accessors_spatial_others(fps):
    assert eq(fps.B.b, fps.E.t, fps.EH.t, fps.AC.b, fps.DF.t, fps.DI.t, )
    assert eq(fps.D.r, fps.E.l, fps.EF.l, fps.AG.r, fps.BH.l, fps.BI.l, )
    assert eq(fps.H.t, fps.E.b, fps.BE.b, fps.GI.t, fps.DF.b, fps.AF.b, )
    assert eq(fps.F.l, fps.E.r, fps.DE.r, fps.CI.l, fps.BH.r, fps.AH.r, )
    assert eq(fps.E.c, fps.AI.c, fps.DF.c, fps.BH.c, )


def test_vector_accessors_spatial(fps):
    assert eq(fps.AC.lrvec, fps.AF.lrvec, fps.AI.lrvec, fps.DF.lrvec, fps.DI.lrvec, fps.GI.lrvec, )
    assert eq(fps.AG.tbvec, fps.AH.tbvec, fps.AI.tbvec, fps.BH.tbvec, fps.BI.tbvec, fps.CI.tbvec, )
    assert eq(fps.AE.diagvec, fps.BF.diagvec, fps.DH.diagvec, fps.EI.diagvec, )
    assert eqall([fp.pxvec for fp in fps.values()] +
                 [fp.diagvec / fp.rsize for fp in fps.values()],)


def test_coordinates_accessors_raster_corners(fps):
    assert eq(fps.AC.rtr, fps.AF.rtr, fps.AI.rtr, fps.DF.rtr, fps.DI.rtr, fps.GI.rtr, )
    assert eq(fps.AG.rbl, fps.AH.rbl, fps.AI.rbl, fps.BH.rbl, fps.BI.rbl, fps.CI.rbl, )
    assert eq(fps.AE.rbr, fps.BF.rbr, fps.DH.rbr, fps.EI.rbr, )


def test_coordinates_accessors_raster_others(fps):
    assert eq(fps.AC.rt, fps.AF.rt, fps.AI.rt, fps.DF.rt, fps.DI.rt, fps.GI.rt, )
    assert eq(fps.AG.rl, fps.AH.rl, fps.AI.rl, fps.BH.rl, fps.BI.rl, fps.CI.rl, )
    assert eq(fps.AE.rr, fps.BF.rr, fps.DH.rr, fps.EI.rr, )
    assert eq(fps.AE.rb, fps.BF.rb, fps.DH.rb, fps.EI.rb, )
    assert eq(fps.AE.rc, fps.BF.rc, fps.DH.rc, fps.EI.rc, )


def test_resolution_and_rotation_accessors(fps):

    c, a, b, f, d, e = fps.A.gt
    assert eq(
        (a, b, c, d, e, f),
        fps.A.aff6,
        fps.A.aff23.flatten(),
        fps.A.aff33.flatten()[:6],
        fps.A.affine[0:6],
    )


    def _of_all(fn):
        return [fn(fp) for fp in fps.values()]

    assert eqall(
        _of_all(lambda fp: fp.gt[1]) +
        _of_all(lambda fp: fp.scale[0]) +
        _of_all(lambda fp: fp.pxsize[0]) +
        _of_all(lambda fp: fp.pxsize[1]) +
        _of_all(lambda fp: fp.pxsizex) +
        _of_all(lambda fp: fp.pxsizey) +
        _of_all(lambda fp: fp.pxvec[0]) +
        _of_all(lambda fp: -fp.gt[5]) +
        _of_all(lambda fp: -fp.scale[1]) +
        _of_all(lambda fp: -fp.pxvec[1])
    )
    assert eqall(
        _of_all(lambda fp: fp.angle) +
        [0]
    )
    assert eqall(_of_all(lambda fp: fp.gt[2]))
    assert eqall(_of_all(lambda fp: fp.gt[4]))
    for fp in fps.values():
        assert eq(fp.gt[0], fp.tlx)
        assert eq(fp.gt[3], fp.tly)


def test_non_trivial_accessors(fps):
    assert eq(fps.AI.semimajoraxis, fps.AH.semimajoraxis, fps.AG.semimajoraxis, )
    assert eq(fps.BH.semiminoraxis, fps.BE.semiminoraxis, fps.B.semiminoraxis, )

    assert eq(
        fps.A.length + fps.B.length + fps.D.length + fps.E.length,
        fps.AE.length + fps.A.length * 2
    )

    assert eq(fps.AI.rsemimajoraxis, fps.AH.rsemimajoraxis, fps.AG.rsemimajoraxis, )
    assert eq(fps.BH.rsemiminoraxis, fps.BE.rsemiminoraxis, fps.B.rsemiminoraxis, )

    assert eq(fps.AI.rarea, np.prod(fps.AI.rsize), np.prod(fps.AI.size / fps.AI.pxsize))

    assert eq(fps.AI.rlength,
              # sum of sides minus 4
              fps.AI.rsizex * 2 + fps.AI.rsizey * 2 - 4,
              # sum of smaller rlength plus delta
              fps.AC.rlength + fps.AD.rsizey * 2,)
    fp = buzz.Footprint(gt=fps.AI.gt, rsize=(2, 10))
    assert eq(fp.rsemiminoraxis, 1)
    assert eq(fp.rlength, fp.rsemimajoraxis * 4)

    fp = buzz.Footprint(gt=fps.AI.gt, rsize=(1, 10))
    assert eq(fp.rsemiminoraxis, 1, tol=1)
    assert eq(fp.rlength, fp.rsemimajoraxis * 2)

    fp = buzz.Footprint(gt=fps.AI.gt, rsize=(1, 1))
    assert eq(fp.rsemiminoraxis, 1, fp.rsemimajoraxis, tol=1)
    assert eq(fp.rlength, 1)

def test_equal(fps):
    dfs = [
        fps.DF,
        fps.AF.intersection(fps.DI),
        fps.DF.intersection(fps.AI),
        fps.DF.intersection(fps.AF),
        fps.DF.intersection(fps.DI),
        fps.DF & fps.DI,
    ]
    for a in dfs:
        assert a == dfs[0]
    bhs = [
        fps.BH,
        fps.AH.intersection(fps.BI),
        fps.BH.intersection(fps.AI),
        fps.BH.intersection(fps.AH),
        fps.BH.intersection(fps.BI),
    ]
    for a in bhs:
        assert a == bhs[0]
    for a, b in itertools.combinations(fps.values(), 2):
        assert a != b


def test_morpho(fps):

    def create(rsizex, rsizey):
        return buzz.Footprint(gt=fps.AI.gt, rsize=(rsizex, rsizey))
    assert eq(create(3, 3).erode(1).rarea, 1 * 1)
    assert eq(create(4, 4).erode(1).rarea, 2 * 2)
    assert eq(create(5, 5).erode(1).rarea, 3 * 3)
    assert eq(create(5, 5).erode(2).rarea, 1 * 1)

    assert eq(create(3, 4).erode(1).rarea, 1 * 2)
    assert eq(create(4, 3).erode(1).rarea, 1 * 2)

    assert eq(create(2, 2).dilate(1).rarea, 4 * 4)
    assert eq(create(2, 3).dilate(1).rarea, 4 * 5)

    assert eq(create(1, 1).dilate(1).dilate(2).erode(3).rarea, 1 * 1)


def test_init_edge_cases(fps):
    # rotation
    aff = Affine.translation(42, 21) * Affine.rotation(12) * Affine.scale(0.1, -0.1)
    with buzz.Env(allow_complex_footprint=True):
        buzz.Footprint(gt=aff.to_gdal(), rsize=[1, 1])

    # missing parameters
    with pytest.raises(ValueError):
        buzz.Footprint(tl=fps.A.tl, size=fps.A.size)
    with pytest.raises(ValueError):
        buzz.Footprint(rsize=fps.A.rsize, size=fps.A.size)
    with pytest.raises(ValueError):
        buzz.Footprint(rsize=fps.A.rsize, tl=fps.A.tl)
    with pytest.raises(ValueError):
        buzz.Footprint(rsize=fps.A.rsize)

    # shapes
    with pytest.raises(ValueError, match='shape'):
        buzz.Footprint(rsize=[], tl=fps.A.tl, size=fps.A.size)
    with pytest.raises(ValueError, match='shape'):
        buzz.Footprint(rsize=fps.A.rsize, tl=[], size=fps.A.size)
    with pytest.raises(ValueError, match='shape'):
        buzz.Footprint(rsize=fps.A.rsize, tl=fps.A.tl, size=[])
    with pytest.raises(ValueError, match='shape'):
        buzz.Footprint(rsize=fps.A.rsize, gt=[])

    # finitude
    with pytest.raises(ValueError):
        buzz.Footprint(rsize=[-1] * 2, tl=fps.A.tl, size=fps.A.size)
    with pytest.raises(ValueError):
        buzz.Footprint(rsize=fps.A.rsize, tl=[np.inf] * 2, size=fps.A.size)
    with pytest.raises(ValueError):
        buzz.Footprint(rsize=fps.A.rsize, tl=fps.A.tl, size=[np.inf] * 2)
    with pytest.raises(ValueError):
        buzz.Footprint(rsize=fps.A.rsize, gt=[np.inf] * 6)

def test_clip(fps1px):
    fps = fps1px
    assert fpeq(
        fps.E,
        fps.E.clip(0, 0, 1, 1),
        fps.E.clip(-1, -1, 1, 1),
        fps.AI.clip(1, 1, 2, 2),
        fps.AI.clip(-2, -2, -1, -1),
        fps.BI.clip(0, 1, 1, 2),
        fps.BI.clip(0 - 2, 1 - 3, 1 - 2, 2 - 3),
    )


def test_move(fps1px):
    fps = fps1px

    with buzz.Env(warnings=False, allow_complex_footprint=True): # Test the `warnings` deprecation
        assert fpeq(
            fps.B,
            fps.A.move(fps.B.tl),
            fps.B.move(fps.B.tl),
            fps.C.move(fps.B.tl),
            fps.A.move(fps.B.tl, fps.B.tr),
            fps.B.move(fps.B.tl, fps.B.tr),
            fps.C.move(fps.B.tl, fps.B.tr),
            fps.A.move(fps.B.tl, fps.B.tr, fps.B.br),
            fps.B.move(fps.B.tl, fps.B.tr, fps.B.br),
            fps.C.move(fps.B.tl, fps.B.tr, fps.B.br),
        )

        aff = (
            Affine.translation(*fps.A.bl) * Affine.rotation(45) * Affine.scale(2**0.5, 2**0.5 * -2)
        )
        assert fpeq(
            buzz.Footprint(gt=aff.to_gdal(), rsize=(1, 1)),
            fps.A.move(fps.A.bl, fps.A.tr, fps.I.tr),
            fps.B.move(fps.A.bl, fps.A.tr, fps.I.tr),
            fps.C.move(fps.A.bl, fps.A.tr, fps.I.tr),
        )
        with pytest.raises(ValueError, match='angle'):
            fps.C.move(fps.A.bl, fps.A.tr, fps.I.c)


def test_binary_predicates(fps):
    for fp in fps.values():
        assert fp.share_area(fps.AI)
        assert fp.same_grid(fps.AI)
        sq2 = 2 ** 0.5
        assert not fp.same_grid(fp.move([sq2, sq2]))
        with buzz.Env(allow_complex_footprint=True):
            assert not fp.same_grid(fp.move([sq2, sq2], [2 * sq2, 2 * sq2]))


def test_numpy_like_functions(fps, fps1px):

    assert eq(
        fps.A.meshgrid_raster,
        fps.B.meshgrid_raster,
        fps.C.meshgrid_raster,
        np.meshgrid(range(fps.A.rw), range(fps.A.rh)),
    )
    assert eq(
        fps.A.meshgrid_spatial,
        fps.C.meshgrid_spatial + fps.A.tl[:, None, None] - fps.C.tl[:, None, None],
        fps.I.meshgrid_spatial + fps.A.tl[:, None, None] - fps.I.tl[:, None, None],
    )

    assert eq(
        fps.A.meshgrid_raster,
        fps.A.meshgrid_raster_in(fps.AI),
        fps.A.meshgrid_raster_in(fps.A),
        fps.I.meshgrid_raster_in(fps.A) + ((fps.A.tl - fps.I.tl) / fps.A.scale)[:, None, None],
    )
    assert fps.A.meshgrid_raster_in(fps.A, dtype='uint8')[0].dtype == np.uint8
    assert fps.A.meshgrid_raster_in(fps.A, dtype='float64', op=42)[0].dtype == np.float64


    fps = fps1px

    dense = fps.AI.move(fps.A.tl, fps.A.tr)
    mesh = np.dstack(fps.EI.meshgrid_raster_in(fps.AI))
    mesh_dense = np.dstack(fps.EI.meshgrid_raster_in(dense))
    assert ((mesh_dense / mesh) == 3).all()

    op = lambda arr: (arr * 0 + 42)

    assert (np.asarray(fps.AI.meshgrid_raster_in(fps.AI, op=op)) == 42).all()

    mg_ai = np.asarray(fps.AI.meshgrid_spatial)
    for fp in fps.values():
        mg = np.asarray(fp.meshgrid_spatial)
        slices = (slice(0, 2),) + fp.slice_in(fps.AI)
        assert (mg == mg_ai[slices]).all()

    assert fps.A.meshgrid_spatial[0][fps.I.slice_in(fps.A, clip=True)].size == 0

def test_coord_conv(fps):
    ai = np.dstack(fps.AI.meshgrid_spatial)

    assert fps.AI.raster_to_spatial(ai).shape == ai.shape
    assert fps.AI.spatial_to_raster(ai).shape == ai.shape
    assert fps.AI.spatial_to_raster(ai, dtype='float16').dtype == np.float16
    assert fps.AI.spatial_to_raster(ai, dtype='float16', op=42).dtype == np.float16



from buzzard._a_source import ASource, ABackSource

class AStored(ASource):
    """Base abstract class defining the common behavior of all sources that are stored somewhere
    (like RAM or disk).

    Features Defined
    ----------------
    - Has an opening mode
    """

    @property
    def mode(self):
        """Open mode, one of {'r', 'w'}"""
        return self._back.mode

class ABackStored(ABackSource):
    """Implementation of AStored's specifications"""

    def __init__(self, mode, **kwargs):
        self.mode = mode
        super(ABackStored, self).__init__(**kwargs)
